from functools import lru_cache

try:
    import mecab_ko as MeCab
    import mecab_ko_dic
except ImportError:
    # Don't fail until the tokenizer is actually used
    MeCab = None

from .tokenizer_base import BaseTokenizer

FAIL_MESSAGE = """
Korean tokenization requires extra dependencies, but you do not have them installed.
Please install them like so.

    pip install sacrebleu[ko]
"""


class TokenizerKoMecab(BaseTokenizer):
    def __init__(self):
        if MeCab is None:
            raise RuntimeError(FAIL_MESSAGE)
        self.tagger = MeCab.Tagger(mecab_ko_dic.MECAB_ARGS + " -Owakati")

        # make sure the dictionary is mecab-ko-dic
        d = self.tagger.dictionary_info()
        assert d.size == 811795, \
            "Please make sure to use the mecab-ko-dic for MeCab-ko"
        # This asserts that no user dictionary has been loaded
        assert d.next is None

    @lru_cache(maxsize=2**16)
    def __call__(self, line):
        """
        Tokenizes an Korean input line using MeCab-ko morphological analyzer.

        :param line: a segment to tokenize
        :return: the tokenized line
        """
        line = line.strip()
        sentence = self.tagger.parse(line).strip()
        return sentence

    def signature(self):
        """
        Returns the MeCab-ko parameters.

        :return: signature string
        """
        signature = self.tagger.version() + "-KO"
        return 'ko-mecab-' + signature



import os

from ..utils import smart_open
from .base import Dataset


class TSVDataset(Dataset):
    """
    The format used by the MTNT datasets. Data is in a single TSV file.
    """

    @staticmethod
    def _split_index_and_filename(meta, field):
        """
        Splits the index and filename from a metadata string.

        e.g. meta="3:en-de.tsv", filed=[Any value] -> (3, "en-de.tsv")
             "en-de.tsv", filed="src" -> (1, "en-de.tsv")
             "en-de.tsv", filed="tgt" -> (2, "en-de.tsv")
        """
        arr = meta.split(":")
        if len(arr) == 2:
            try:
                index = int(arr[0])
            except ValueError:
                raise Exception(f"Invalid meta for TSVDataset: {meta}")
            return index, arr[1]

        else:
            index = 0 if field == "src" else 1
            return index, meta

    def process_to_text(self, langpair=None):
        """Processes raw files to plain text files.

        :param langpair: The language pair to process. e.g. "en-de". If None, all files will be processed.
        """
        # ensure that the dataset is downloaded
        self.maybe_download()
        langpairs = self._get_langpair_metadata(langpair)

        for langpair in langpairs:
            fieldnames = self.fieldnames(langpair)
            origin_files = [
                os.path.join(self._rawdir, path) for path in langpairs[langpair]
            ]

            for field, origin_file, meta in zip(
                fieldnames, origin_files, langpairs[langpair]
            ):
                index, origin_file = self._split_index_and_filename(meta, field)

                origin_file = os.path.join(self._rawdir, origin_file)
                output_file = self._get_txt_file_path(langpair, field)

                with smart_open(origin_file) as fin:
                    with smart_open(output_file, "wt") as fout:
                        for line in fin:
                            # be careful with empty source or reference lines
                            # MTNT2019/ja-en.final.tsv:632 `'1033\t718\t\t\n'`
                            print(line.rstrip("\n").split("\t")[index], file=fout)



# -*- coding: utf-8 -*-

import os
import logging

from functools import lru_cache
from ..utils import SACREBLEU_DIR, download_file
from .tokenizer_base import BaseTokenizer

sacrelogger = logging.getLogger('sacrebleu')


SPM_MODELS = {
    "spm": {
        "url": "https://dl.fbaipublicfiles.com/fairseq/models/flores/sacrebleu_tokenizer_spm.model",
        "signature": "flores101",
    },
    # same as the default of "spm"
    "flores101": {
        "url": "https://dl.fbaipublicfiles.com/fairseq/models/flores/sacrebleu_tokenizer_spm.model",
        "signature": "flores101",
    },
    "flores200": {
        "url": "https://tinyurl.com/flores200sacrebleuspm",
        "signature": "flores200",
    },
}

class TokenizerSPM(BaseTokenizer):
    def signature(self):
        return self.name

    def __init__(self, key="spm"):
        self.name = SPM_MODELS[key]["signature"]

        if key == "spm":
            sacrelogger.warn("Tokenizer 'spm' has been changed to 'flores101', and may be removed in the future.")

        try:
            import sentencepiece as spm
        except (ImportError, ModuleNotFoundError):
            raise ImportError(
                '\n\nPlease install the sentencepiece library for SPM tokenization:'
                '\n\n  pip install sentencepiece '
            )
        self.sp = spm.SentencePieceProcessor()

        model_path = os.path.join(SACREBLEU_DIR, "models", os.path.basename(SPM_MODELS[key]["url"]))
        if not os.path.exists(model_path):
            url = SPM_MODELS[self.name]["url"]
            download_file(url, model_path)
        self.sp.Load(model_path)

    @lru_cache(maxsize=2**16)
    def __call__(self, line):
        """Tokenizes all the characters in the input line.

        :param line: a segment to tokenize
        :return: the tokenized line
        """
        return " ".join(self.sp.EncodeAsPieces(line))


class Flores200Tokenizer(TokenizerSPM):
    def __init__(self):
        super().__init__("flores200")

class Flores101Tokenizer(TokenizerSPM):
    def __init__(self):
        super().__init__("flores101")



"""The implementation of the TER metric (Snover et al., 2006)."""

# Copyright 2020 Memsource
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import List, Dict, Sequence, Optional, Any

from ..tokenizers.tokenizer_ter import TercomTokenizer
from ..utils import sum_of_lists
from .base import Score, Signature, Metric
from .lib_ter import translation_edit_rate


class TERSignature(Signature):
    """A convenience class to represent the reproducibility signature for TER.

    :param args: key-value dictionary passed from the actual metric instance.
    """
    def __init__(self, args: dict):
        """`TERSignature` initializer."""
        super().__init__(args)
        self._abbr.update({
            'case': 'c',
            'tok': 't',
            'norm': 'nr',
            'punct': 'pn',
            'asian': 'as',
        })

        self.info.update({
            'case': 'mixed' if args['case_sensitive'] else 'lc',
            'tok': args['tokenizer_signature'],
            'norm': args['normalized'],
            'punct': not args['no_punct'],
            'asian': args['asian_support'],
        })


class TERScore(Score):
    """A convenience class to represent TER scores.

    :param score: The TER score.
    :param num_edits: The cumulative number of edits.
    :param ref_length: The cumulative average reference length.
    """
    def __init__(self, score: float, num_edits: float, ref_length: float):
        """`TERScore` initializer."""
        super().__init__('TER', score)
        self.num_edits = int(num_edits)
        self.ref_length = ref_length


class TER(Metric):
    """Translation edit rate (TER). A near-exact reimplementation of the Tercom
    algorithm, produces identical results on all "sane" outputs.

    Tercom original implementation: https://github.com/jhclark/tercom

    The beam edit distance algorithm uses a slightly different approach (we stay
    around the diagonal which is faster, at least in Python) so in some
    (extreme) corner cases, the output could differ.

    Caching in the edit distance is based partly on the PyTer package by Hiroyuki
    Tanaka (MIT license). (https://github.com/aflc/pyter)

    :param normalized: Enable character normalization. By default, normalizes a couple of things such as
        newlines being stripped, retrieving XML encoded characters, and fixing tokenization for punctuation. When
        'asian_support' is enabled, also normalizes specific Asian (CJK) character sequences, i.e.
        split them down to the character level.
    :param no_punct: Remove punctuation. Can be used in conjunction with 'asian_support' to also remove typical
        punctuation markers in Asian languages (CJK).
    :param asian_support: Enable special treatment of Asian characters. This option only has an effect when
        'normalized' and/or 'no_punct' is enabled. If 'normalized' is also enabled, then Asian (CJK)
        characters are split down to the character level. If 'no_punct' is enabled alongside 'asian_support',
        specific unicode ranges for CJK and full-width punctuations are also removed.
    :param case_sensitive: If `True`, does not lowercase sentences.
    :param references: A sequence of reference documents with document being
        defined as a sequence of reference strings. If given, the reference info
        will be pre-computed and cached for faster re-computation across many systems.
    """

    _SIGNATURE_TYPE = TERSignature

    def __init__(self, normalized: bool = False,
                 no_punct: bool = False,
                 asian_support: bool = False,
                 case_sensitive: bool = False,
                 references: Optional[Sequence[Sequence[str]]] = None):
        """`TER` initializer."""
        super().__init__()

        self.no_punct = no_punct
        self.normalized = normalized
        self.asian_support = asian_support
        self.case_sensitive = case_sensitive

        self.tokenizer = TercomTokenizer(
            normalized=self.normalized,
            no_punct=self.no_punct,
            asian_support=self.asian_support,
            case_sensitive=self.case_sensitive,
        )
        self.tokenizer_signature = self.tokenizer.signature()

        if references is not None:
            self._ref_cache = self._cache_references(references)

    def _preprocess_segment(self, sent: str) -> str:
        """Given a sentence, apply tokenization if enabled.

        :param sent: The input sentence string.
        :return: The pre-processed output string.
        """
        return self.tokenizer(sent.rstrip())

    def _compute_score_from_stats(self, stats: List[float]) -> TERScore:
        """Computes the final score from already aggregated statistics.

        :param stats: A list or numpy array of segment-level statistics.
        :return: A `TERScore` object.
        """
        total_edits, sum_ref_lengths = stats[0], stats[1]

        if sum_ref_lengths > 0:
            score = total_edits / sum_ref_lengths
        elif total_edits > 0:
            score = 1.0  # empty reference(s) and non-empty hypothesis
        else:
            score = 0.0  # both reference(s) and hypothesis are empty

        return TERScore(100 * score, total_edits, sum_ref_lengths)

    def _aggregate_and_compute(self, stats: List[List[float]]) -> TERScore:
        """Computes the final TER score given the pre-computed corpus statistics.

        :param stats: A list of segment-level statistics
        :return: A `TERScore` instance.
        """
        return self._compute_score_from_stats(sum_of_lists(stats))

    def _compute_segment_statistics(
            self, hypothesis: str, ref_kwargs: Dict) -> List[float]:
        """Given a (pre-processed) hypothesis sentence and already computed
        reference words, returns the segment statistics required to compute
        the full TER score.

        :param hypothesis: Hypothesis sentence.
        :param ref_kwargs: A dictionary with `ref_words` key which is a list
        where each sublist contains reference words.
        :return: A two-element list that contains the 'minimum number of edits'
        and 'the average reference length'.
        """

        ref_lengths = 0
        best_num_edits = int(1e16)

        words_hyp = hypothesis.split()

        # Iterate the references
        ref_words = ref_kwargs['ref_words']
        for words_ref in ref_words:
            num_edits, ref_len = translation_edit_rate(words_hyp, words_ref)
            ref_lengths += ref_len
            if num_edits < best_num_edits:
                best_num_edits = num_edits

        avg_ref_len = ref_lengths / len(ref_words)
        return [best_num_edits, avg_ref_len]

    def _extract_reference_info(self, refs: Sequence[str]) -> Dict[str, Any]:
        """Given a list of reference segments, applies pre-processing & tokenization
        and returns list of tokens for each reference.

        :param refs: A sequence of strings.
        :return: A dictionary that will be passed to `_compute_segment_statistics()`
        through keyword arguments.
        """
        ref_words = []

        for ref in refs:
            ref_words.append(self._preprocess_segment(ref).split())

        return {'ref_words': ref_words}



import os

from ..utils import smart_open
from .base import Dataset


class PlainTextDataset(Dataset):
    """
    The plain text format. Data is separated into source and reference files.
    Each line of the two files is aligned.
    """

    def process_to_text(self, langpair=None):
        """Processes raw files to plain text files.

        :param langpair: The language pair to process. e.g. "en-de". If None, all files will be processed.
        """
        # ensure that the dataset is downloaded
        self.maybe_download()
        langpairs = self._get_langpair_metadata(langpair)

        for langpair in langpairs:
            fieldnames = self.fieldnames(langpair)
            origin_files = [
                os.path.join(self._rawdir, path) for path in langpairs[langpair]
            ]

            for field, origin_file in zip(fieldnames, origin_files):

                origin_file = os.path.join(self._rawdir, origin_file)
                output_file = self._get_txt_file_path(langpair, field)

                with smart_open(origin_file) as fin:
                    with smart_open(output_file, "wt") as fout:
                        for line in fin:
                            print(line.rstrip(), file=fout)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2017--2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

__description__ = "Hassle-free computation of shareable, comparable, and reproducible BLEU, chrF, and TER scores"


# Backward compatibility functions for old style API access (<= 1.4.10)
from .compat import (
    corpus_bleu,
    corpus_chrf,
    corpus_ter,
    raw_corpus_bleu,
    sentence_bleu,
    sentence_chrf,
    sentence_ter,
)
from .dataset import DATASETS
from .metrics import BLEU, CHRF, TER
from .metrics.helpers import extract_char_ngrams, extract_word_ngrams
from .utils import (
    SACREBLEU_DIR,
    download_test_set,
    get_available_testsets,
    get_langpairs_for_testset,
    get_reference_files,
    get_source_file,
    smart_open,
)
from .version import __version__

__all__ = [
    "smart_open",
    "SACREBLEU_DIR",
    "download_test_set",
    "get_source_file",
    "get_reference_files",
    "get_available_testsets",
    "get_langpairs_for_testset",
    "extract_word_ngrams",
    "extract_char_ngrams",
    "DATASETS",
    "BLEU",
    "CHRF",
    "TER",
    "corpus_bleu",
    "raw_corpus_bleu",
    "sentence_bleu",
    "corpus_chrf",
    "sentence_chrf",
    "corpus_ter",
    "sentence_ter",
    "__version__",
]



from functools import lru_cache
from .tokenizer_base import BaseTokenizer


class TokenizerChar(BaseTokenizer):
    def signature(self):
        return 'char'

    def __init__(self):
        pass

    @lru_cache(maxsize=2**16)
    def __call__(self, line):
        """Tokenizes all the characters in the input line.

        :param line: a segment to tokenize
        :return: the tokenized line
        """
        return " ".join((char for char in line))



import os

import lxml.etree as ET

from ..utils import smart_open
from .base import Dataset

from collections import defaultdict


def _get_field_by_translator(translator):
    if not translator:
        return "ref"
    else:
        return f"ref:{translator}"

class WMTXMLDataset(Dataset):
    """
    The 2021+ WMT dataset format. Everything is contained in a single file.
    Can be parsed with the lxml parser.
    """
    @staticmethod
    def _unwrap_wmt21_or_later(raw_file):
        """
        Unwraps the XML file from wmt21 or later.
        This script is adapted from https://github.com/wmt-conference/wmt-format-tools

        :param raw_file: The raw xml file to unwrap.
        :return: Dictionary which contains the following fields
            (each a list with values for each sentence):
            - `src`: The source sentences.
            - `docid`: ID indicating which document the sentences belong to.
            - `origlang`: The original language of the document.
            - `domain`: Domain of the document.
            - `ref:{translator}`: The references produced by each translator.
            - `ref`: An alias for the references from the first translator.
        """
        tree = ET.parse(raw_file)
        # Find and check the documents (src, ref, hyp)
        src_langs, ref_langs, translators = set(), set(), set()
        for src_doc in tree.getroot().findall(".//src"):
            src_langs.add(src_doc.get("lang"))

        for ref_doc in tree.getroot().findall(".//ref"):
            ref_langs.add(ref_doc.get("lang"))
            translator = ref_doc.get("translator")
            translators.add(translator)

        assert (
            len(src_langs) == 1
        ), f"Multiple source languages found in the file: {raw_file}"
        assert (
            len(ref_langs) == 1
        ), f"Found {len(ref_langs)} reference languages found in the file: {raw_file}"

        src = []
        docids = []
        orig_langs = []
        domains = []

        refs = { _get_field_by_translator(translator): [] for translator in translators }

        systems = defaultdict(list)

        src_sent_count, doc_count, seen_domain = 0, 0, False
        for doc in tree.getroot().findall(".//doc"):
            # Skip the testsuite
            if "testsuite" in doc.attrib:
                continue

            doc_count += 1
            src_sents = {
                int(seg.get("id")): seg.text for seg in doc.findall(".//src//seg")
            }

            def get_sents(doc):
                return {
                    int(seg.get("id")): seg.text if seg.text else ""
                    for seg in doc.findall(".//seg")
                }

            ref_docs = doc.findall(".//ref")

            trans_to_ref = {
                ref_doc.get("translator"): get_sents(ref_doc) for ref_doc in ref_docs
            }

            hyp_docs = doc.findall(".//hyp")
            hyps = {
                hyp_doc.get("system"): get_sents(hyp_doc) for hyp_doc in hyp_docs
            }

            for seg_id in sorted(src_sents.keys()):
                # no ref translation is available for this segment
                if not any([value.get(seg_id, "") for value in trans_to_ref.values()]):
                    continue
                for translator in translators:
                    refs[_get_field_by_translator(translator)].append(
                        trans_to_ref.get(translator, {translator: {}}).get(seg_id, "")
                    )
                src.append(src_sents[seg_id])
                for system_name in hyps.keys():
                    systems[system_name].append(hyps[system_name][seg_id])
                docids.append(doc.attrib["id"])
                orig_langs.append(doc.attrib["origlang"])
                # The "domain" attribute is missing in WMT21 and WMT22
                domains.append(doc.get("domain"))
                seen_domain = doc.get("domain") is not None
                src_sent_count += 1

        fields = {"src": src, **refs, "docid": docids, "origlang": orig_langs, **systems}
        if seen_domain:
            fields["domain"] = domains
        return fields

    def _get_langpair_path(self, langpair):
        """
        Returns the path for this language pair.
        This is useful because in WMT22, the language-pair data structure can be a dict,
        in order to allow for overriding which test set to use.
        """
        langpair_data = self._get_langpair_metadata(langpair)[langpair]
        rel_path = langpair_data["path"] if isinstance(langpair_data, dict) else langpair_data[0]
        return os.path.join(self._rawdir, rel_path)

    def process_to_text(self, langpair=None):
        """Processes raw files to plain text files.

        :param langpair: The language pair to process. e.g. "en-de". If None, all files will be processed.
        """
        # ensure that the dataset is downloaded
        self.maybe_download()

        for langpair in sorted(self._get_langpair_metadata(langpair).keys()):
            # The data type can be a list of paths, or a dict, containing the "path"
            # and an override on which labeled reference to use (key "refs")
            rawfile = self._get_langpair_path(langpair)

            with smart_open(rawfile) as fin:
                fields = self._unwrap_wmt21_or_later(fin)

            for fieldname in fields:
                textfile = self._get_txt_file_path(langpair, fieldname)

                # skip if the file already exists
                if os.path.exists(textfile) and os.path.getsize(textfile) > 0:
                    continue

                with smart_open(textfile, "w") as fout:
                    for line in fields[fieldname]:
                        print(self._clean(line), file=fout)

    def _get_langpair_allowed_refs(self, langpair):
        """
        Returns the preferred references for this language pair.
        This can be set in the language pair block (as in WMT22), and backs off to the
        test-set-level default, or nothing.

        There is one exception. In the metadata, sometimes there is no translator field
        listed (e.g., wmt22:liv-en). In this case, the reference is set to "", and the
        field "ref" is returned.
        """
        defaults = self.kwargs.get("refs", [])
        langpair_data = self._get_langpair_metadata(langpair)[langpair]
        if isinstance(langpair_data, dict):
            allowed_refs = langpair_data.get("refs", defaults)
        else:
            allowed_refs = defaults
        allowed_refs = [_get_field_by_translator(ref) for ref in allowed_refs]

        return allowed_refs

    def get_reference_files(self, langpair):
        """
        Returns the requested reference files.
        This is defined as a default at the test-set level, and can be overridden per language.
        """
        # Iterate through the (label, file path) pairs, looking for permitted labels
        allowed_refs = self._get_langpair_allowed_refs(langpair)
        all_files = self.get_files(langpair)
        all_fields = self.fieldnames(langpair)
        ref_files = [
            f for f, field in zip(all_files, all_fields) if field in allowed_refs
        ]
        return ref_files

    def fieldnames(self, langpair):
        """
        Return a list of all the field names. For most source, this is just
        the source and the reference. For others, it might include the document
        ID for each line, or the original language (origLang).

        get_files() should return the same number of items as this.

        :param langpair: The language pair (e.g., "de-en")
        :return: a list of field names
        """
        self.maybe_download()
        rawfile = self._get_langpair_path(langpair)

        with smart_open(rawfile) as fin:
            fields = self._unwrap_wmt21_or_later(fin)

        return list(fields.keys())



import os
import logging
import multiprocessing as mp
from typing import Sequence, Dict, Optional, Tuple, List, Union, Any, Mapping

import numpy as np

from .metrics.base import Metric, Score, Signature

IS_WINDOWS = os.name == 'nt'


sacrelogger = logging.getLogger('sacrebleu')


class Result:
    """A container to represent results from a particular statistical
    significance test.
    :param score: The floating point score for the system at hand.
    :param p_value: If exists, represents the p-value when the system at
    hand is compared to a baseline using a paired test.
    :param mean: When paired bootstrap test is applied, this represents
    the true mean score estimated from bootstrap resamples of the system.
    :param ci: When paired bootstrap test is applied, this represents
    the 95% confidence interval around the true mean score `sys_mean`.
    """
    def __init__(self, score: float, p_value: Optional[float] = None,
                 mean: Optional[float] = None, ci: Optional[float] = None):
        self.score = score
        self.p_value = p_value
        self.mean = mean
        self.ci = ci

    def __repr__(self):
        return ','.join([f'{k}={str(v)}' for k, v in self.__dict__.items()])


def estimate_ci(scores: np.ndarray) -> Tuple[float, float]:
    """Takes a list of scores and returns mean and 95% confidence
    interval around the mean.

    :param scores: A list of floating point scores.
    :return: A tuple of mean and the 95% CI.
    """
    # Sort the scores
    scores = np.sort(scores)
    n = len(scores)

    # Get CI bounds (95%, i.e. 1/40 from left)
    lower_idx = n // 40
    upper_idx = n - lower_idx - 1
    lower, upper = scores[lower_idx], scores[upper_idx]
    ci = 0.5 * (upper - lower)
    return (scores.mean(), ci)


def _bootstrap_resample(stats: List[List[Union[int, float]]],
                        metric: Metric, n_samples: int = 1000) -> Tuple[str, List[Score]]:
    """Performs bootstrap resampling for a single system to estimate
    a confidence interval around the true mean.
    :param stats: A list of statistics extracted from the system's hypotheses.
    :param metric: The `Metric` instance to be used for score computation.
    :n_samples: Number of bootstrap resamples to use.

    :return: A tuple of the seed choice as string and the list of `Score`
    instances for all bootstrap resamples.
    """

    # Set numpy RNG's seed
    # If  given -> Fix to the given value
    # If given but =='[Nn]one', don't fix the seed i.e. pull entropy from OS
    seed = os.environ.get('SACREBLEU_SEED', '12345')
    _seed = None if seed.lower() == 'none' else int(seed)
    rng = np.random.default_rng(_seed)

    # The indices that'll produce all bootstrap resamples at once
    idxs = rng.choice(len(stats), size=(n_samples, len(stats)), replace=True)

    # convert to numpy array. float32 is more efficient
    stats_np = np.array(stats, dtype='float32')

    # recompute scores for all resamples
    scores = [
        metric._compute_score_from_stats(_s.sum(0)) for _s in stats_np[idxs]]

    return str(seed).lower(), scores


def _compute_p_value(stats: np.ndarray, real_difference: float) -> float:
    """Computes the p-value given the sample statistics and the real statistic.
    :param stats: A numpy array with the sample statistics.
    :real_difference: The real statistic.
    :return: The p-value.
    """
    # Taken from: significance/StratifiedApproximateRandomizationTest.java
    # https://github.com/jhclark/multeval.git

    # "the != is important. if we want to score the same system against itself
    # having a zero difference should not be attributed to chance."

    c = np.sum(stats > real_difference).item()

    # "+1 applies here, though it only matters for small numbers of shufflings,
    # which we typically never do. it's necessary to ensure the probability of
    # falsely rejecting the null hypothesis is no greater than the rejection
    # level of the test (see william and morgan on significance tests)
    p = (c + 1) / (len(stats) + 1)

    return p


def _paired_ar_test(baseline_info: Dict[str, Tuple[np.ndarray, Result]],
                    sys_name: str,
                    hypotheses: Sequence[str],
                    references: Optional[Sequence[Sequence[str]]],
                    metrics: Dict[str, Metric],
                    n_samples: int = 10000,
                    n_ar_confidence: int = -1,
                    seed: Optional[int] = None) -> Tuple[str, Dict[str, Result]]:
    """Paired two-sided approximate randomization (AR) test for MT evaluation.

    :param baseline_info: A dictionary with `Metric` instances as the keys,
    that contains sufficient statistics and a `Result` instance for the baseline system.
    :param sys_name: The name of the system to be evaluated.
    :param hypotheses: A sequence of string hypotheses for the system.
    :param references: A sequence of reference documents with document being
    defined as a sequence of reference strings. If `None`, references
    will be used through each metric's internal cache.
    :param metrics: A dictionary of `Metric` instances that will be computed
    for each system.
    :param n_samples: The number of AR trials.
    :param n_ar_confidence: The number of bootstrap resamples to use for
    confidence estimation. A value of -1 disables confidence estimation.
    :param seed: The seed value for the RNG. If `None`, the RNG will not be
    fixed to a particular seed.

    :return: A tuple with first element being the system name and the second
    being a `Result` namedtuple.
    """
    # Seed the RNG
    rng = np.random.default_rng(seed)

    # Generate indices that'll select stats
    pos_sel = rng.integers(2, size=(n_samples, len(hypotheses)), dtype=bool)

    # Flip mask to obtain selectors for system hypotheses
    neg_sel = ~pos_sel

    if n_ar_confidence > 0:
        # Perform confidence estimation as well
        bs_idxs = rng.choice(
            len(hypotheses), size=(n_ar_confidence, len(hypotheses)), replace=True)

    results = {}

    for name, metric in metrics.items():
        # Use pre-computed match stats for the baseline
        bl_stats, bl_result = baseline_info[name]

        # Compute system's stats and score
        sacrelogger.info(f'Computing {name} for {sys_name!r} and extracting sufficient statistics')
        sys_stats = metric._extract_corpus_statistics(hypotheses, references)
        sys_score = metric._aggregate_and_compute(sys_stats)

        # original test statistic: absolute difference between baseline and the system
        diff = abs(bl_result.score - sys_score.score)

        sacrelogger.info(f' > Performing approximate randomization test (# trials: {n_samples})')
        # get shuffled pseudo systems
        shuf_a = pos_sel @ bl_stats + neg_sel @ sys_stats
        shuf_b = neg_sel @ bl_stats + pos_sel @ sys_stats

        # Aggregate trial stats and compute scores for each
        scores_a = np.array(
            [metric._aggregate_and_compute(x).score for x in shuf_a[:, None]])
        scores_b = np.array(
            [metric._aggregate_and_compute(x).score for x in shuf_b[:, None]])

        # Count the statistical difference and compute the p-value
        p = _compute_p_value(
            np.abs(np.array(scores_a) - np.array(scores_b)), diff)

        res = Result(sys_score.score, p)

        if n_ar_confidence > 0:
            sacrelogger.info(f' > Performing bootstrap resampling for confidence interval (# resamples: {n_ar_confidence})')
            sys_stats = np.array(sys_stats, dtype='float32')
            # recompute scores for all resamples
            sys_scores = np.array([
                metric._compute_score_from_stats(_s.sum(0)).score for _s in sys_stats[bs_idxs]
            ])
            res.mean, res.ci = estimate_ci(sys_scores)

        # Store the result
        results[name] = res

    return sys_name, results


def _paired_bs_test(baseline_info: Dict[str, Tuple[np.ndarray, Result]],
                    sys_name: str,
                    hypotheses: Sequence[str],
                    references: Optional[Sequence[Sequence[str]]],
                    metrics: Dict[str, Metric],
                    n_samples: int = 1000,
                    n_ar_confidence: int = -1,
                    seed: Optional[int] = None) -> Tuple[str, Dict[str, Result]]:
    """Paired bootstrap resampling test for MT evaluation. This function
    replicates the behavior of the Moses script called
    `bootstrap-hypothesis-difference-significance.pl`.

    :param baseline_info: A dictionary with `Metric` instances as the keys,
    that contains sufficient statistics and a `Result` instance for the baseline system.
    :param sys_name: The name of the system to be evaluated.
    :param hypotheses: A sequence of string hypotheses for the system.
    :param references: A sequence of reference documents with document being
    defined as a sequence of reference strings. If `None`, references
    will be used through each metric's internal cache.
    :param metrics: A dictionary of `Metric` instances that will be computed
    for each system.
    :param n_samples: The number of bootstrap resamples.
    :param n_ar_confidence: This parameter is not used for this function but
    is there for signature compatibility in the API.
    :param seed: The seed value for the RNG. If `None`, the RNG will not be
    fixed to a particular seed.

    :return: A tuple with first element being the system name and the second
    being a `Result` namedtuple.
    """
    # Seed the RNG
    rng = np.random.default_rng(seed)

    results = {}

    # It takes ~10ms to generated the indices
    idxs = rng.choice(
        len(hypotheses), size=(n_samples, len(hypotheses)), replace=True)

    for name, metric in metrics.items():
        # Use pre-computed match stats for the baseline
        bl_stats, bl_result = baseline_info[name]

        # Compute system's stats and score
        sacrelogger.info(f'Computing {name} for {sys_name!r} and extracting sufficient statistics')
        sys_stats = metric._extract_corpus_statistics(hypotheses, references)
        sys_score = metric._aggregate_and_compute(sys_stats)

        # Convert to numpy arrays for efficient indexing
        sys_stats = np.array(sys_stats, dtype='float32')
        bl_stats = np.array(bl_stats, dtype='float32')

        # original test statistic: absolute difference between baseline and the system
        diff = abs(bl_result.score - sys_score.score)

        sacrelogger.info(f' > Performing paired bootstrap resampling test (# resamples: {n_samples})')
        scores_bl = np.array(
            [metric._compute_score_from_stats(_s.sum(0)).score for _s in bl_stats[idxs]])
        scores_sys = np.array(
            [metric._compute_score_from_stats(_s.sum(0)).score for _s in sys_stats[idxs]])

        # Compute CI as well
        sys_mean, sys_ci = estimate_ci(scores_sys)

        # Compute the statistics
        sample_diffs = np.abs(scores_sys - scores_bl)
        stats = sample_diffs - sample_diffs.mean()

        # Count the statistical difference and compute the p-value
        p = _compute_p_value(stats, diff)

        results[name] = Result(sys_score.score, p, sys_mean, sys_ci)

    return sys_name, results


class PairedTest:
    """This is the manager class that will call the actual standalone implementation
    for approximate randomization or paired bootstrap resampling, based on the
    `test_type` argument.

    :param named_systems: A lisf of (system_name, system_hypotheses) tuples on
    which the test will be applied.
    :param metrics: A dictionary of `Metric` instances that will be computed
    for each system.
    :param references: A sequence of reference documents with document being
    defined as a sequence of reference strings. If `None`, already cached references
    will be used through each metric's internal cache.
    :param test_type: `ar` for approximate randomization, `bs` for paired bootstrap.
    :param n_samples: The number of AR trials (for `ar`) or bootstrap resamples (for `bs`).
    The defaults (10000 or 1000 respectively) will be used if 0 is passed.
    :param n_ar_confidence: If `approximate randomization` is selected, the number
    of bootstrap resamples to use for confidence estimation. A value of -1 disables
    confidence estimation. 0 will use the default of 1000.
    :param n_jobs: If 0, a worker process will be spawned for each system variant.
    If > 0, the number of workers will be set accordingly. The default of 1
    does not use multi-processing.
    """
    _DEFAULT_SAMPLES = {
        'ar': 10000,
        'bs': 1000,
    }

    def __init__(self, named_systems: List[Tuple[str, Sequence[str]]],
                 metrics: Mapping[str, Metric],
                 references: Optional[Sequence[Sequence[str]]],
                 test_type: str = 'ar',
                 n_samples: int = 0,
                 n_ar_confidence: int = -1,
                 n_jobs: int = 1):
        assert test_type in ('ar', 'bs'), f"Unknown test type {test_type!r}"
        self.test_type = test_type

        # Set method
        if self.test_type == 'ar':
            self._fn = _paired_ar_test
        elif self.test_type == 'bs':
            self._fn = _paired_bs_test

        # Set numpy RNG's seed
        # If  given -> Fix to the given value
        # If given but =='[Nn]one', don't fix the seed i.e. pull entropy from OS
        seed = os.environ.get('SACREBLEU_SEED', '12345')
        self._seed = None if seed.lower() == 'none' else int(seed)
        self.n_jobs = n_jobs
        self.references = references
        self.named_systems = named_systems

        # Set the defaults if requested
        self.n_ar_confidence = n_ar_confidence if n_ar_confidence != 0 else \
            self._DEFAULT_SAMPLES['bs']

        self.n_samples = n_samples if n_samples > 0 else \
            self._DEFAULT_SAMPLES[self.test_type]

        # Number of systems (excluding the baseline)
        self.n_systems = len(named_systems) - 1

        # Decide on number of workers
        if IS_WINDOWS:
            sacrelogger.warning('Parallel tests are not supported on Windows.')
            self.n_jobs = 1
        elif self.n_jobs == 0:
            # Decide automatically
            # Divide by two to ignore hyper-threading
            n_max_jobs = mp.cpu_count() // 2
            if n_max_jobs == 0:
                self.n_jobs = 1
            else:
                # Don't use more workers than the number of CPUs
                self.n_jobs = min(n_max_jobs, self.n_systems)

        self._signatures: Dict[str, Signature] = {}
        self._baseline_info: Dict[str, Tuple[Any, Result]] = {}

        ##################################################
        # Pre-compute and cache baseline system statistics
        ##################################################
        self.metrics = {}

        bl_name, bl_hyps = self.named_systems[0]

        for name, metric in metrics.items():
            sacrelogger.info(f'Pre-computing {name} statistics for {bl_name!r}')
            bl_stats = metric._extract_corpus_statistics(bl_hyps, self.references)
            bl_score = metric._aggregate_and_compute(bl_stats)

            # Compute CI for the baseline here once
            confidence_n = self.n_samples if self.test_type == 'bs' \
                else self.n_ar_confidence

            bl_mean, bl_ci = None, None
            if confidence_n > 0:
                _, bl_scores = _bootstrap_resample(bl_stats, metric, confidence_n)
                bl_mean, bl_ci = estimate_ci(np.array([x.score for x in bl_scores]))

            result = Result(bl_score.score, mean=bl_mean, ci=bl_ci)
            # Use updated name for the metric
            self._baseline_info[bl_score.name] = (bl_stats, result)
            self.metrics[bl_score.name] = metric

            # Update metric signature as well
            sig = metric.get_signature()
            sig.update('seed', str(self._seed).lower())

            # Num samples for bs, num trials for AR
            sig.update(self.test_type, self.n_samples)
            if self.n_ar_confidence > 0:
                # Bootstrap is used for AR CI as well
                sig.update('bs', self.n_ar_confidence)
            self._signatures[bl_score.name] = sig

    def __call__(self) -> Tuple[Dict[str, Signature], Dict[str, List[Union[str, Result]]]]:
        """Runs the paired test either on single or multiple worker processes."""
        tasks = []
        scores: Dict[str, List[Union[str, Result]]] = {}

        # Add the name column
        scores['System'] = [ns[0] for ns in self.named_systems]

        # Store baseline results as the first position
        for metric, (_, result) in self._baseline_info.items():
            scores[metric] = [result]

        # Prepare list of arguments for each comparison
        # Skip the baseline (pos: 0)
        for idx, (name, hyps) in enumerate(self.named_systems[1:]):
            seed = self._seed if self._seed else None

            tasks.append(
                (self._baseline_info, name, hyps, self.references,
                 self.metrics, self.n_samples, self.n_ar_confidence, seed))

        # Run the test(s)
        if self.n_jobs == 1:
            results = [self._fn(*args) for args in tasks]
        else:
            # NOTE: The overhead of worker creation is not negligible
            # but if you have many systems and TER enabled, this significantly
            # speeds up the test.
            # NOTE: This only works on Linux/Mac OS X but not Windows. Windows only
            # supports `spawn` backend which requires things to be called
            # from within __main__.
            sacrelogger.info(f'Launching {self.n_jobs} parallel workers.')
            with mp.get_context('fork').Pool(self.n_jobs) as pool:
                jobs = [pool.apply_async(self._fn, args) for args in tasks]

                # wait for completion
                results = [j.get() for j in jobs]

        # Keep the order deterministic
        for sys_name, sys_results in results:
            for metric, _result in sys_results.items():
                scores[metric].append(_result)

        return self._signatures, scores



"""The base `Score`, `Metric` and `Signature` classes to derive from.

`Metric` is an abstract class that enforces the implementation of a set
of abstract methods. This way, a correctly implemented metric will work
seamlessly with the rest of the codebase.
"""

import json
import logging
import statistics
from abc import ABCMeta, abstractmethod
from typing import Any, Dict, List, Optional, Sequence

from ..version import __version__

sacrelogger = logging.getLogger("sacrebleu")


class Score:
    """A base score class to derive from.

    :param name: The name of the underlying metric.
    :param score: A floating point number for the final metric.
    """

    def __init__(self, name: str, score: float):
        """`Score` initializer."""
        self.name = name
        self.score = score

        # Statistical test related fields
        self._mean = -1.0
        self._ci = -1.0

        # More info can be added right after the score
        self._verbose = ""

    def format(
        self,
        width: int = 2,
        score_only: bool = False,
        signature: str = "",
        is_json: bool = False,
    ) -> str:
        """Returns a pretty representation of the score.
        :param width: Floating point decimal precision width.
        :param score_only: If `True`, and the format is not `json`,
        returns a single score string.
        :param signature: A string representation of the given `Signature`
        instance.
        :param is_json: If `True`, will output the score in JSON string.
        :return: A plain or JSON-formatted string representation.
        """
        d = {
            "name": self.name,
            "score": float(f"{self.score:.{width}f}"),
            "signature": signature,
        }

        sc = f"{self.score:.{width}f}"

        if self._mean > 0:
            confidence_mean = f"{self._mean:.{width}f}"
            confidence_var = f"{self._ci:.{width}f}"
            confidence_str = f"μ = {confidence_mean} ± {confidence_var}"

            sc += f" ({confidence_str})"
            if is_json:
                d["confidence_mean"] = float(confidence_mean)
                d["confidence_var"] = float(confidence_var)
                d["confidence"] = confidence_str

        # Construct full score line
        full_score = f"{self.name}|{signature}" if signature else self.name
        full_score = f"{full_score} = {sc}"
        if self._verbose:
            full_score += f" {self._verbose}"
            d["verbose_score"] = self._verbose

        if score_only:
            return sc

        if is_json:
            for param in signature.split("|"):
                key, value = param.split(":")
                d[key] = value
            return json.dumps(d, indent=1, ensure_ascii=False)

        return full_score

    def estimate_ci(self, scores: List["Score"]):
        """Takes a list of scores and stores mean, stdev and 95% confidence
        interval around the mean.

        :param scores: A list of `Score` objects obtained from bootstrap
        resampling for example.
        """
        # Sort the scores
        raw_scores = sorted([x.score for x in scores])
        n = len(raw_scores)

        # Get CI bounds (95%, i.e. 1/40 from left)
        lower_idx = n // 40
        upper_idx = n - lower_idx - 1
        lower, upper = raw_scores[lower_idx], raw_scores[upper_idx]
        self._ci = 0.5 * (upper - lower)
        self._mean = statistics.mean(raw_scores)

    def __repr__(self):
        """Returns a human readable score string."""
        return self.format()


class Signature:
    """A convenience class to represent sacreBLEU reproducibility signatures.

    :param args: key-value dictionary passed from the actual metric instance.
    """

    def __init__(self, args: dict):
        """`Signature` initializer."""
        # Global items that are shared across all metrics
        self._abbr = {
            "version": "v",
            "nrefs": "#",
            "test": "t",
            "lang": "l",
            "subset": "S",
            "origlang": "o",
            "bs": "bs",  # Bootstrap resampling trials
            "ar": "ar",  # Approximate randomization trials
            "seed": "rs",  # RNG's seed
        }

        if "num_refs" not in args:
            raise ValueError(
                "Number of references unknown, please evaluate the metric first."
            )

        num_refs = args["num_refs"]
        if num_refs == -1:
            # Detect variable number of refs
            num_refs = "var"

        # Global items that are shared across all metrics
        # None's will be ignored
        self.info = {
            "version": __version__,
            "nrefs": num_refs,
            "bs": args.get("n_bootstrap", None),
            "ar": None,
            "seed": args.get("seed", None),
            "test": args.get("test_set", None),
            "lang": args.get("langpair", None),
            "origlang": args.get("origlang", None),
            "subset": args.get("subset", None),
        }

    def format(self, short: bool = False) -> str:
        """Returns a string representation of the signature.

        :param short: If True, shortened signature is produced.
        :return: A string representation of the signature.
        """
        pairs = []
        keys = list(self.info.keys())
        # keep version always at end
        keys.remove("version")
        for name in keys + ["version"]:
            value = self.info[name]
            if value is not None:
                if isinstance(value, bool):
                    # Replace True/False with yes/no
                    value = "yes" if value else "no"
                final_name = self._abbr[name] if short else name
                pairs.append(f"{final_name}:{value}")

        return "|".join(pairs)

    def update(self, key: str, value: Any):
        """Add a new item or update an existing one.

        :param key: The key to use in the dictionary.
        :param value: The associated value for the `key`.
        """
        self.info[key] = value

    def __str__(self):
        """Returns a human-readable signature string."""
        return self.format()

    def __repr__(self):
        """Returns a human-readable signature string."""
        return self.format()


class Metric(metaclass=ABCMeta):
    """A base class for all metrics that ensures the implementation of some
    methods. Much of the common functionality is moved to this base class
    from other metrics."""

    # Each metric should define its Signature class' name here
    _SIGNATURE_TYPE = Signature

    def __init__(self):
        """`Metric` initializer."""
        # The pre-computed reference cache
        self._ref_cache = None

        # only useful for BLEU tokenized warnings. Set to True so that
        # warnings are not issued for other metrics.
        self._force = True

        # Will be used by the signature when bootstrap resampling
        self.n_bootstrap = None
        self.seed = None

    def _check_sentence_score_args(self, hyp: str, refs: Sequence[str]):
        """Performs sanity checks on `sentence_score` method's arguments.

        :param hyp: A single hypothesis string.
        :param refs: A sequence of reference strings.
        """
        prefix = self.__class__.__name__
        err_msg = None

        if not isinstance(hyp, str):
            err_msg = "The argument `hyp` should be a string."
        elif isinstance(refs, str) or not isinstance(refs, Sequence):
            err_msg = "The argument `refs` should be a sequence of strings."
        elif not isinstance(refs[0], str) and refs[0] is not None:
            err_msg = "Each element of `refs` should be a string."

        if err_msg:
            raise TypeError(f"{prefix}: {err_msg}")

    def _check_corpus_score_args(
        self, hyps: Sequence[str], refs: Optional[Sequence[Sequence[str]]]
    ):
        """Performs sanity checks on `corpus_score` method's arguments.

        :param hypses: A sequence of hypothesis strings.
        :param refs: A sequence of reference documents with document being
        defined as a sequence of reference strings. If `None`, cached references
        will be used.
        """

        prefix = self.__class__.__name__
        err_msg = None

        if not isinstance(hyps, Sequence):
            err_msg = "`hyps` should be a sequence of strings."
        elif not isinstance(hyps[0], str):
            err_msg = "Each element of `hyps` should be a string."
        elif any(line is None for line in hyps):
            err_msg = "Undefined line in hypotheses stream!"

        if refs is not None:
            if not isinstance(refs, Sequence):
                err_msg = "`refs` should be a sequence of sequence of strings."
            elif not isinstance(refs[0], Sequence):
                err_msg = "Each element of `refs` should be a sequence of strings."
            elif not isinstance(refs[0][0], str) and refs[0][0] is not None:
                err_msg = "`refs` should be a sequence of sequence of strings."

        if err_msg:
            raise TypeError(f"{prefix}: {err_msg}")

    @abstractmethod
    def _aggregate_and_compute(self, stats: List[List[Any]]) -> Any:
        """Computes the final score given the pre-computed match statistics.

        :param stats: A list of segment-level statistics.
        :return: A `Score` instance.
        """
        pass

    @abstractmethod
    def _compute_score_from_stats(self, stats: List[Any]) -> Any:
        """Computes the final score from already aggregated statistics.

        :param stats: A list or numpy array of segment-level statistics.
        :return: A `Score` object.
        """
        pass

    @abstractmethod
    def _preprocess_segment(self, sent: str) -> str:
        """A wrapper around the metric's tokenization and pre-processing logic.
        This should be implemented for reference caching to work correctly.

        :param sent: The input sentence.
        :return: The pre-processed output sentence.
        """
        pass

    @abstractmethod
    def _extract_reference_info(self, refs: Sequence[str]) -> Dict[str, Any]:
        """Given a list of reference segments, extract the required
        information (such as n-grams for BLEU and chrF). This should be implemented
        for the generic `_cache_references()` to work across all metrics.

        :param refs: A sequence of strings.
        """
        pass

    @abstractmethod
    def _compute_segment_statistics(
        self, hypothesis: str, ref_kwargs: Dict
    ) -> List[Any]:
        """Given a (pre-processed) hypothesis sentence and already computed
        reference info, returns the best match statistics across the
        references. The return type is usually a List of ints or floats.

        :param hypothesis: A pre-processed hypothesis sentence.
        :param ref_kwargs: A dictionary with reference-related information
        within. This is formulated as a dictionary as different metrics may
        require different information regarding a reference segment.
        """
        pass

    def _cache_references(self, references: Sequence[Sequence[str]]) -> List[Any]:
        """Given the full set of document references, extract segment n-grams
        (or other necessary information) for caching purposes.

        :param references: A sequence of reference documents with document being
        defined as a sequence of reference strings. A particular reference
        segment can be '' or `None` to allow the use of variable number
        of references per segment.
        :return: A list where each element is a tuple of segment n-grams and
        reference lengths, as returned by `_extract_reference_info()`.
        """
        ref_cache = []

        # Decide on final number of refs here as well
        num_refs = set()

        for refs in zip(*references):
            # Remove undefined references
            lines = [x for x in refs if x is not None]

            # Keep track of reference counts to allow variable reference
            # info in the signature
            num_refs.add(len(lines))

            lines = [self._preprocess_segment(x) for x in lines]

            # Get n-grams
            ref_cache.append(self._extract_reference_info(lines))

        if len(num_refs) == 1:
            self.num_refs = list(num_refs)[0]
        else:
            # A variable number of refs exist
            self.num_refs = -1

        return ref_cache

    def _extract_corpus_statistics(
        self, hypotheses: Sequence[str], references: Optional[Sequence[Sequence[str]]]
    ) -> Any:
        """Reads the corpus and returns sentence-level match statistics for
        faster re-computations esp. during statistical tests.

        :param hypotheses: A sequence of hypothesis strings.
        :param references: A sequence of reference documents with document being
        defined as a sequence of reference strings. If `None`, cached references
        will be used.
        :return: A list where each sublist corresponds to segment statistics.
        """
        # Pre-compute references
        # Don't store the cache as the user is explicitly passing refs
        if references:
            ref_cache = self._cache_references(references)
        elif self._ref_cache:
            ref_cache = self._ref_cache
        else:
            raise RuntimeError("No references provided and the cache is empty.")

        stats = []
        tok_count = 0

        for hyp, ref_kwargs in zip(hypotheses, ref_cache):
            # Check for already-tokenized input problem (only for BLEU)
            if not self._force and hyp.endswith(" ."):
                tok_count += 1

            hyp = self._preprocess_segment(hyp)

            # Collect stats
            stats.append(self._compute_segment_statistics(hyp, ref_kwargs))

        if tok_count >= 100:
            sacrelogger.warning("That's 100 lines that end in a tokenized period ('.')")
            sacrelogger.warning(
                "It looks like you forgot to detokenize your test data, which may hurt your score."
            )
            sacrelogger.warning(
                "If you insist your data is detokenized, or don't care, you can suppress this message with the `force` parameter."
            )

        return stats

    def sentence_score(self, hypothesis: str, references: Sequence[str]) -> Any:
        """Compute the metric for a single sentence against a single (or multiple) reference(s).

        :param hypothesis: A single hypothesis string.
        :param references: A sequence of reference strings.
        :return: A `Score` object.
        """
        self._check_sentence_score_args(hypothesis, references)

        stats = self._extract_corpus_statistics(
            [hypothesis], [[refs] for refs in references]
        )
        return self._aggregate_and_compute(stats)

    def corpus_score(
        self,
        hypotheses: Sequence[str],
        references: Optional[Sequence[Sequence[str]]],
        n_bootstrap: int = 1,
    ) -> Any:
        """Compute the metric for a corpus against a single (or multiple) reference(s).

        :param hypotheses: A sequence of hypothesis strings.
        :param references: A sequence of reference documents with document being
        defined as a sequence of reference strings. If `None`, cached references
        will be used.
        :param n_bootstrap: If > 1, provides 95% confidence interval around true mean
        using bootstrap resampling with `n_bootstrap` samples.
        :return: A `Score` object.
        """
        self._check_corpus_score_args(hypotheses, references)

        # Collect corpus stats
        stats = self._extract_corpus_statistics(hypotheses, references)

        # Compute the actual system score
        actual_score = self._aggregate_and_compute(stats)

        if n_bootstrap > 1:
            # Compute bootstrap estimate as well
            # Delayed import is to escape from numpy import if bootstrap
            # is not requested.
            from ..significance import _bootstrap_resample

            self.n_bootstrap = n_bootstrap
            self.seed, bs_scores = _bootstrap_resample(stats, self, n_bootstrap)
            actual_score.estimate_ci(bs_scores)

        return actual_score

    def get_signature(self) -> Signature:
        """Creates and returns the signature for the metric. The creation
        of signatures is delayed as the number of references is resolved
        only at the point of reference caching."""
        return self._SIGNATURE_TYPE(self.__dict__)



#!/usr/bin/env python3

# Copyright 2017--2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

"""
SacreBLEU provides hassle-free computation of shareable, comparable, and reproducible BLEU scores.
Inspired by Rico Sennrich's `multi-bleu-detok.perl`, it produces the official WMT scores but works with plain text.
It also knows all the standard test sets and handles downloading, processing, and tokenization for you.

See the [README.md] file for more information.
"""

import io
import os
import sys
import logging
import pathlib
import argparse
from collections import defaultdict


# Allows calling the script as a standalone utility
# See: https://github.com/mjpost/sacrebleu/issues/86
if __package__ is None and __name__ == '__main__':
    parent = pathlib.Path(__file__).absolute().parents[1]
    sys.path.insert(0, str(parent))
    __package__ = 'sacrebleu'

from .dataset import DATASETS
from .metrics import METRICS
from .utils import smart_open, filter_subset, get_langpairs_for_testset, get_available_testsets
from .utils import print_test_set, print_subset_results, get_reference_files, download_test_set
from .utils import args_to_dict, sanity_check_lengths, print_results_table, print_single_results
from .utils import get_available_testsets_for_langpair, Color

from . import __version__ as VERSION

sacrelogger = logging.getLogger('sacrebleu')

try:
    # SIGPIPE is not available on Windows machines, throwing an exception.
    from signal import SIGPIPE  # type: ignore

    # If SIGPIPE is available, change behaviour to default instead of ignore.
    from signal import signal, SIG_DFL
    signal(SIGPIPE, SIG_DFL)
except ImportError:
    pass


def parse_args():
    arg_parser = argparse.ArgumentParser(
        description='sacreBLEU: Hassle-free computation of shareable BLEU scores.\n'
                    'Quick usage: score your detokenized output against WMT\'14 EN-DE:\n'
                    '    cat output.detok.de | sacrebleu -t wmt14 -l en-de',
        formatter_class=argparse.RawDescriptionHelpFormatter)

    arg_parser.add_argument('--citation', '--cite', default=False, action='store_true',
                            help='Dump the bibtex citation and quit.')
    arg_parser.add_argument('--list', default=False, action='store_true',
                            help='Print a list of all available test sets.')
    arg_parser.add_argument('--test-set', '-t', type=str, default=None,
                            help='The test set to use (see also --list) or a comma-separated list of test sets to be concatenated.')
    arg_parser.add_argument('--language-pair', '-l', dest='langpair', default=None,
                            help='Source-target language pair (2-char ISO639-1 codes).')
    arg_parser.add_argument('--origlang', '-ol', dest='origlang', default=None,
                            help='Use a subset of sentences with a given original language (2-char ISO639-1 codes), "non-" prefix means negation.')
    arg_parser.add_argument('--subset', dest='subset', default=None,
                            help='Use a subset of sentences whose document annotation matches a given regex (see SUBSETS in the source code).')
    arg_parser.add_argument('--download', type=str, default=None,
                            help='Download a test set and quit.')
    arg_parser.add_argument('--echo', nargs="+", type=str, default=None,
                            help='Output the source (src), reference (ref), or other available field (docid, ref:A, ref:1 for example) to STDOUT and quit. '
                                 'You can get available fields with options `--list` and `-t`' 'For example: `sacrebleu -t wmt21 --list`. '
                                 'If multiple fields are given, they are outputted with tsv format in the order they are given.'
                                 'You can also use `--echo all` to output all available fields.')

    # I/O related arguments
    # Multiple input files can be provided for significance testing for example
    arg_parser.add_argument('--input', '-i', type=str, nargs='*', default=None,
                            help='Read input from file(s) instead of STDIN.')
    arg_parser.add_argument('refs', nargs='*', default=[],
                            help='Optional list of references. If given, it should preceed the -i/--input argument.')
    arg_parser.add_argument('--num-refs', '-nr', type=int, default=1,
                            help='Split the reference stream on tabs, and expect this many references. (Default: %(default)s)')
    arg_parser.add_argument('--encoding', '-e', type=str, default='utf-8',
                            help='Open text files with specified encoding (Default: %(default)s)')

    # Metric selection
    avail_metrics = [m.lower() for m in METRICS]
    arg_parser.add_argument('--metrics', '-m', choices=avail_metrics, nargs='+', default=['bleu'],
                            help='Space-delimited list of metrics to compute (Default: bleu)')
    arg_parser.add_argument('--sentence-level', '-sl', action='store_true', help='Compute metric for each sentence.')

    # BLEU-related arguments
    # since sacreBLEU had only support for BLEU initially, the argument names
    # are not prefixed with 'bleu' as in chrF arguments for example.
    # Let's do that manually here through dest= options, as otherwise
    # things will get quite hard to maintain when other metrics are added.
    bleu_args = arg_parser.add_argument_group('BLEU related arguments')

    bleu_args.add_argument('--smooth-method', '-s', choices=METRICS['BLEU'].SMOOTH_DEFAULTS.keys(), default='exp',
                           dest='bleu_smooth_method',
                           help='Smoothing method: exponential decay, floor (increment zero counts), add-k (increment num/denom by k for n>1), or none. (Default: %(default)s)')
    bleu_args.add_argument('--smooth-value', '-sv', type=float, default=None,
                           dest='bleu_smooth_value',
                           help='The smoothing value. Only valid for floor and add-k. '
                                f"(Defaults: floor: {METRICS['BLEU'].SMOOTH_DEFAULTS['floor']}, "
                                f"add-k: {METRICS['BLEU'].SMOOTH_DEFAULTS['add-k']})")
    bleu_args.add_argument('--tokenize', '-tok', choices=METRICS['BLEU'].TOKENIZERS, default=None,
                           dest='bleu_tokenize',
                           help='Tokenization method to use for BLEU. If not provided, defaults to `zh` for Chinese, '
                                '`ja-mecab` for Japanese, `ko-mecab` for Korean and `13a` (mteval) otherwise.')
    bleu_args.add_argument('--lowercase', '-lc', dest='bleu_lowercase', action='store_true', default=False,
                           help='If True, enables case-insensitivity. (Default: %(default)s)')
    bleu_args.add_argument('--force', default=False, action='store_true',
                           dest='bleu_force', help='Insist that your tokenized input is actually detokenized.')

    # ChrF-related arguments
    chrf_args = arg_parser.add_argument_group('chrF related arguments')
    chrf_args.add_argument('--chrf-char-order', '-cc', type=int, default=METRICS['CHRF'].CHAR_ORDER,
                           help='Character n-gram order. (Default: %(default)s)')
    chrf_args.add_argument('--chrf-word-order', '-cw', type=int, default=METRICS['CHRF'].WORD_ORDER,
                           help='Word n-gram order (Default: %(default)s). If equals to 2, the metric is referred to as chrF++.')
    chrf_args.add_argument('--chrf-beta', type=int, default=METRICS['CHRF'].BETA,
                           help='Determine the importance of recall w.r.t precision. (Default: %(default)s)')
    chrf_args.add_argument('--chrf-whitespace', action='store_true', default=False,
                           help='Include whitespaces when extracting character n-grams. (Default: %(default)s)')
    chrf_args.add_argument('--chrf-lowercase', action='store_true', default=False,
                           help='Enable case-insensitivity. (Default: %(default)s)')
    chrf_args.add_argument('--chrf-eps-smoothing', action='store_true', default=False,
                           help='Enables epsilon smoothing similar to chrF++.py, NLTK and Moses; instead of effective order smoothing. (Default: %(default)s)')

    # TER related arguments
    ter_args = arg_parser.add_argument_group("TER related arguments (The defaults replicate TERCOM's behavior)")
    ter_args.add_argument('--ter-case-sensitive', action='store_true',
                          help='Enables case sensitivity. (Default: %(default)s)')
    ter_args.add_argument('--ter-asian-support', action='store_true',
                          help='Enables special treatment of Asian characters. (Default: %(default)s)')
    ter_args.add_argument('--ter-no-punct', action='store_true',
                          help='Removes punctuation. (Default: %(default)s)')
    ter_args.add_argument('--ter-normalized', action='store_true',
                          help='Applies basic normalization and tokenization. (Default: %(default)s)')

    # Bootstrap resampling for confidence intervals
    sign_args = arg_parser.add_argument_group('Confidence interval (CI) estimation for single-system evaluation')
    sign_args.add_argument('--confidence', '-ci', action='store_true',
                           help='Report confidence interval using bootstrap resampling.')
    sign_args.add_argument('--confidence-n', '-cin', type=int, default=1000,
                           help='Set the number of bootstrap resamples for CI estimation (Default: %(default)s).')

    # Paired significance testing
    pair_args = arg_parser.add_argument_group('Paired significance testing for multi-system evaluation')
    pair_args_choice = pair_args.add_mutually_exclusive_group()

    pair_args_choice.add_argument('--paired-ar', '-par', action='store_true',
                                  help='Perform paired test using approximate randomization (AR). This option is '
                                       'mutually exclusive with --paired-bs (Default: %(default)s).')
    pair_args_choice.add_argument('--paired-bs', '-pbs', action='store_true',
                                  help='Perform paired test using bootstrap resampling. This option is '
                                       'mutually exclusive with --paired-ar (Default: %(default)s).')

    pair_args.add_argument('--paired-ar-n', '-parn', type=int, default=10000,
                           help='Number of trials for approximate randomization test (Default: %(default)s).')

    pair_args.add_argument('--paired-bs-n', '-pbsn', type=int, default=1000,
                           help='Number of bootstrap resamples for paired bootstrap resampling test (Default: %(default)s).')

    pair_args.add_argument('--paired-jobs', '-j', type=int, default=1,
                           help='If 0, launches as many workers as the number of systems. If > 0, sets the number of workers manually. '
                                'This feature is currently not supported on Windows.')

    # Reporting related arguments
    report_args = arg_parser.add_argument_group('Reporting related arguments')
    report_args.add_argument('--quiet', '-q', default=False, action='store_true',
                             help='Suppress verbose messages.')
    report_args.add_argument('--short', '-sh', default=False, action='store_true',
                             help='Produce a shorter (less human readable) signature.')
    report_args.add_argument('--score-only', '-b', default=False, action='store_true',
                             help='Print only the computed score.')
    report_args.add_argument('--width', '-w', type=int, default=1,
                             help='Floating point width (Default: %(default)s).')
    report_args.add_argument('--detail', '-d', default=False, action='store_true',
                             help='Print detailed information (split test sets based on origlang).')
    report_args.add_argument('--no-color', '-nc', action='store_true',
                             help='Disable the occasional use of terminal colors.')

    output_formats = ['json', 'text', 'latex']
    report_args.add_argument('--format', '-f', default='json', choices=output_formats,
                             help='Set the output format. `latex` is only valid for multi-system mode whereas '
                                  '`json` and `text` apply to single-system mode only. This flag is overridden if the '
                                  'SACREBLEU_FORMAT environment variable is set to one of the valid choices (Default: %(default)s).')

    arg_parser.add_argument('--version', '-V', action='version', version='%(prog)s {}'.format(VERSION))

    args = arg_parser.parse_args()

    # Override the format from the environment, if any
    if 'SACREBLEU_FORMAT' in os.environ:
        _new_value = os.environ['SACREBLEU_FORMAT'].lower()
        if _new_value in output_formats:
            args.format = _new_value

    return args


def main():
    args = parse_args()

    # Is paired test requested?
    paired_test_mode = args.paired_bs or args.paired_ar

    # Explicitly set the encoding
    sys.stdin = open(sys.stdin.fileno(), mode='r', encoding='utf-8', buffering=True, newline="\n")
    sys.stdout = open(sys.stdout.fileno(), mode='w', encoding='utf-8', buffering=True)

    if os.environ.get('NO_COLOR', False) or args.no_color:
        Color.ENABLE_COLORS = False
    else:
        # These should come after all stdout manipulations otherwise cause
        # issues esp. on Windows
        import colorama
        colorama.init()

    if not args.quiet:
        logging.basicConfig(level=logging.INFO, format='sacreBLEU: %(message)s')

    if args.download:
        download_test_set(args.download, args.langpair)
        sys.exit(0)

    if args.list:
        if args.test_set:
            for pair in [args.langpair] if args.langpair else get_langpairs_for_testset(args.test_set):
                fields = DATASETS[args.test_set].fieldnames(pair)
                print(f'{pair}: {", ".join(fields)}')
        else:
            if args.langpair:
                print(f'The available test sets for {args.langpair} are:')
                testsets = get_available_testsets_for_langpair(args.langpair)
            else:
                print('The available test sets are:')
                testsets = get_available_testsets()
            for testset in sorted(testsets):
                desc = DATASETS[testset].description.strip()
                print(f'{testset:<30}: {desc}')
        sys.exit(0)

    if args.sentence_level and len(args.metrics) > 1:
        sacrelogger.error('Only one metric can be used in sentence-level mode.')
        sys.exit(1)

    if args.citation:
        if not args.test_set:
            sacrelogger.error('I need a test set (-t).')
            sys.exit(1)
        for test_set in args.test_set.split(','):
            if 'citation' not in DATASETS[test_set]:
                sacrelogger.error(f'No citation found for {test_set}')
            else:
                print(DATASETS[test_set].citation)
        sys.exit(0)

    if args.num_refs != 1 and (args.test_set is not None or len(args.refs) > 1):
        sacrelogger.error('The --num-refs argument allows you to provide any number of tab-delimited references in a single file.')
        sacrelogger.error('You can only use it with externally provided references, however (i.e., not with `-t`),')
        sacrelogger.error('and you cannot then provide multiple reference files.')
        sys.exit(1)

    if args.test_set is not None:
        for test_set in args.test_set.split(','):
            if test_set not in DATASETS:
                sacrelogger.error(f'Unknown test set {test_set!r}')
                sacrelogger.error('Please run with --list to see the available test sets.')
                sys.exit(1)

    if args.test_set is None:
        if len(args.refs) == 0:
            sacrelogger.error('If manual references given, make sure to provide them '
                              'before the -i/--input argument to avoid confusion.')
            sacrelogger.error('Otherwise, I need a predefined test set (-t) from the following list:')
            sacrelogger.error(get_available_testsets())
            sys.exit(1)
    elif len(args.refs) > 0:
        sacrelogger.error('I need exactly one of (a) a predefined test set (-t) or (b) a list of references')
        sys.exit(1)
    elif args.langpair is None:
        sacrelogger.error('I need a language pair (-l). Use --list to see available language pairs for this test set.')
        sys.exit(1)
    else:
        for test_set in args.test_set.split(','):
            langpairs = get_langpairs_for_testset(test_set)
            if args.langpair not in langpairs:
                sacrelogger.error(f'No such language pair {args.langpair!r}')
                sacrelogger.error(f'Available language pairs for {test_set!r} are:')
                for lp in langpairs:
                    sacrelogger.error(f' > {lp}')
                sys.exit(1)

    if args.echo:
        if args.langpair is None or args.test_set is None:
            sacrelogger.warning("--echo requires a test set (--t) and a language pair (-l)")
            sys.exit(1)
        for test_set in args.test_set.split(','):
            print_test_set(test_set, args.langpair, args.echo, args.origlang, args.subset)
        sys.exit(0)

    # Hack: inject target language info for BLEU, so that it can
    # select the tokenizer based on it
    if args.langpair:
        args.bleu_trg_lang = args.langpair.split('-')[1]

    if args.test_set is not None and args.bleu_tokenize == 'none':
        sacrelogger.warning(
            "You are turning off BLEU's internal tokenizer "
            "presumably to supply your own tokenized files.")
        sacrelogger.warning(
            "Published numbers will not be comparable to other papers.")

    # concat_ref_files is a list of list of reference filenames
    # (concatenation happens if multiple test sets are given through -t)
    # Example: [[testset1_refA, testset1_refB], [testset2_refA, testset2_refB]]
    concat_ref_files = []
    if args.test_set is None:
        concat_ref_files.append(args.refs)
    else:
        # Multiple test sets can be given
        for test_set in args.test_set.split(','):
            ref_files = get_reference_files(test_set, args.langpair)
            if len(ref_files) == 0:
                sacrelogger.warning(
                    f'No references found for test set {test_set}/{args.langpair}.')
            concat_ref_files.append(ref_files)

    #################
    # Read references
    #################
    full_refs = [[] for x in range(max(len(concat_ref_files[0]), args.num_refs))]
    for ref_files in concat_ref_files:
        for refno, ref_file in enumerate(ref_files):
            for lineno, line in enumerate(smart_open(ref_file, encoding=args.encoding), 1):
                line = line.rstrip()
                if args.num_refs == 1:
                    full_refs[refno].append(line)
                else:
                    refs = line.split(sep='\t', maxsplit=args.num_refs - 1)
                    # We are strict in fixed number of references through CLI
                    # But the API supports having variable refs per each segment
                    # by simply having '' or None's as dummy placeholders
                    if len(refs) != args.num_refs:
                        sacrelogger.error(f'FATAL: line {lineno}: expected {args.num_refs} fields, but found {len(refs)}.')
                        sys.exit(17)
                    for refno, ref in enumerate(refs):
                        full_refs[refno].append(ref)

    # Decide on the number of final references, override the argument
    args.num_refs = len(full_refs)

    # Read hypotheses
    # Can't tokenize yet as each metric has its own way of tokenizing things
    full_systems, sys_names = [], []

    if args.input is None:
        # Read from STDIN
        inputfh = io.TextIOWrapper(sys.stdin.buffer, encoding=args.encoding)

        # guess the number of systems by looking at the first line
        fields = inputfh.readline().rstrip().split('\t')

        # Set number of systems
        num_sys = len(fields)

        # place the first lines already
        full_systems = [[s] for s in fields]

        # Enumerate the systems
        sys_names = [f'System {i + 1}' for i in range(num_sys)]

        # Read the rest
        for line in inputfh:
            fields = line.rstrip().split('\t')
            if len(fields) != num_sys:
                sacrelogger.error('FATAL: the number of tab-delimited fields in the input stream differ across lines.')
                sys.exit(17)
            # Place systems into the list
            for sys_idx, sent in enumerate(fields):
                full_systems[sys_idx].append(sent.rstrip())
    else:
        # Separate files are given for each system output
        # Ex: --input smt.txt nmt.txt
        for fname in args.input:
            sys_name = fname

            if sys_name in sys_names:
                if paired_test_mode and sys_name == sys_names[0]:
                    # We skip loading a system, if it was already the baseline
                    sacrelogger.info(f'Ignoring {sys_name!r} as it was also given as the baseline.')
                    continue
                else:
                    # To avoid ambiguities, we fail if two systems have same names
                    sacrelogger.error(f"{sys_name!r} already used to name a system.")
                    sacrelogger.error("Make sure to have a different basename for each system.")
                    sys.exit(1)

            # Read the system
            lines = []
            for line in smart_open(fname, encoding=args.encoding):
                lines.append(line.rstrip())
            full_systems.append(lines)
            sys_names.append(sys_name)

        # Set final number of systems
        num_sys = len(sys_names)

    # Add baseline prefix to the first system for clarity
    if paired_test_mode:
        if args.input is None:
            # STDIN mode, no explicit system names
            sys_names = ['Baseline'] + [f'System {i + 1}' for i in range(num_sys - 1)]
        else:
            # --input mode, we have names for the systems, just change the 1st one
            sys_names[0] = f'Baseline: {sys_names[0]}'

    if args.sentence_level:
        if num_sys > 1:
            sacrelogger.error('Only one system can be evaluated in sentence-level mode.')
            sys.exit(1)
        if args.confidence or paired_test_mode:
            sacrelogger.error('Statistical tests are unavailable in sentence-level mode.')
            sys.exit(1)

        # >=2.0.0: effective_order is now part of BLEU class. For sentence-BLEU
        # we now need to explicitly enable it without user's intervention
        # for backward compatibility.
        args.bleu_effective_order = True

    if paired_test_mode and num_sys == 1:
        sacrelogger.error('Paired tests require multiple input systems given to --input (-i).')
        sys.exit(1)

    if num_sys > 1 and args.confidence:
        sacrelogger.error('Use paired tests (--paired) for multiple systems.')
        sys.exit(1)

    # Filter subsets if requested
    outputs = filter_subset(
        [*full_systems, *full_refs], args.test_set, args.langpair,
        args.origlang, args.subset)

    # Unpack systems & references back
    systems, refs = outputs[:num_sys], outputs[num_sys:]

    # Perform some sanity checks
    for system in systems:
        if len(system) == 0:
            message = f'Test set {args.test_set!r} contains no sentence'
            if args.origlang is not None or args.subset is not None:
                message += ' with'
                if args.origlang:
                    message += f' origlang={args.origlang}'
                if args.subset:
                    message += f' subset={args.subset}' + args.subset
            sacrelogger.error(message)
            sys.exit(1)

        # Check lengths
        sanity_check_lengths(system, refs, test_set=args.test_set)

    # Create the metrics
    metrics = {}
    for name in args.metrics:
        # Each metric's specific arguments are prefixed with `metricname_`
        # for grouping. Filter accordingly and strip the prefixes prior to
        # metric object construction.
        metric_args = args_to_dict(args, name.lower(), strip_prefix=True)

        # This will cache reference stats for faster re-computation if required
        metric_args['references'] = refs

        # Make it uppercase for the rest of the code
        name = name.upper()
        metrics[name] = METRICS[name](**metric_args)

    # Handle sentence level and quit
    if args.sentence_level:
        # one metric and one system in use for sentence-level
        metric, system = list(metrics.values())[0], systems[0]

        for hypothesis, *references in zip(system, *refs):
            score = metric.sentence_score(hypothesis, references)
            sig = metric.get_signature().format(args.short)
            print(score.format(args.width, args.score_only, sig))

        sys.exit(0)

    if args.detail and args.format == 'json':
        # The translationese info will interfere with JSON output, disable
        args.format = 'text'

    ##############################
    # Corpus level evaluation mode
    ##############################
    if num_sys == 1:
        # Single system evaluation mode
        results = []
        for name in sorted(metrics):
            # compute the score
            score = metrics[name].corpus_score(
                system, references=None,
                n_bootstrap=args.confidence_n if args.confidence else 1)
            # get the signature
            sig = metrics[name].get_signature().format(
                args.short if args.format != 'json' else False)
            results.append(
                score.format(args.width, args.score_only, sig, args.format == 'json'))

        print_single_results(results, args)

        # Prints detailed information for translationese effect experiments
        if args.detail:
            print_subset_results(metrics, full_systems[0], full_refs, args)
    else:
        # Multi-system evaluation mode
        named_systems = [(sys_names[i], systems[i]) for i in range(num_sys)]
        sacrelogger.info(f'Found {num_sys} systems.')

        if not paired_test_mode:
            # Bootstrap resampling or the usual single score computation mode
            sigs = {}
            scores = defaultdict(list)
            scores['System'] = sys_names

            for sys_name, system in named_systems:
                for name in sorted(metrics):
                    score = metrics[name].corpus_score(system, references=None)
                    sigs[score.name] = metrics[name].get_signature().format(args.short)
                    scores[score.name].append(score.format(args.width, True))

        else:
            # Paired significance testing mode
            from .significance import PairedTest

            # Set params
            test_type = 'bs' if args.paired_bs else 'ar'
            n_samples = args.paired_bs_n if args.paired_bs else args.paired_ar_n

            ps = PairedTest(named_systems, metrics, references=None,
                            test_type=test_type, n_samples=n_samples,
                            n_jobs=args.paired_jobs)

            # Set back the number of trials
            args.paired_n = ps.n_samples

            # Run the test
            sigs, scores = ps()

            # Get signature strings
            sigs = {k: v.format(args.short) for k, v in sigs.items()}

        # Dump the results
        print_results_table(scores, sigs, args)


if __name__ == '__main__':
    main()



import os
import re

from ..utils import smart_open
from .base import Dataset


class FakeSGMLDataset(Dataset):
    """
    The fake SGML format used by WMT prior to 2021. Can't be properly parsed.
    Source and reference(s) in separate files.
    """

    def _convert_format(self, input_file_path, output_filep_path):
        """
        Extract data from raw file and convert to raw txt format.
        """
        with smart_open(input_file_path) as fin, smart_open(
            output_filep_path, "wt"
        ) as fout:
            for line in fin:
                if line.startswith("<seg "):
                    line = self._clean(re.sub(r"<seg.*?>(.*)</seg>.*?", "\\1", line))
                    print(line, file=fout)

    def _convert_meta(self, input_file_path, field, output_filep_path):
        """
        Extract metadata from document tags, projects across segments.
        """
        with smart_open(input_file_path) as fin, smart_open(
            output_filep_path, "wt"
        ) as fout:
            value = ""
            for line in fin:
                if line.startswith("<doc "):
                    match = re.search(rf'{field}="(.*?)"', line)
                    if match is not None:
                        value = match.group(1)

                elif line.startswith("<seg "):
                    # print the current value once for each field
                    print(value, file=fout)

    def process_to_text(self, langpair=None):
        """Processes raw files to plain text files.

        :param langpair: The language pair to process. e.g. "en-de". If None, all files will be processed.
        """
        # ensure that the dataset is downloaded
        self.maybe_download()
        langpairs = self._get_langpair_metadata(langpair)

        for langpair in langpairs:
            fieldnames = self.fieldnames(langpair)
            origin_files = [
                os.path.join(self._rawdir, path) for path in langpairs[langpair]
            ]

            # Add the source file three more times for docid, genre, origlang
            origin_files += [
                os.path.join(self._rawdir, langpairs[langpair][0]) for _ in range(3)
            ]

            for field, origin_file in zip(fieldnames, origin_files):

                origin_file = os.path.join(self._rawdir, origin_file)
                output_file = self._get_txt_file_path(langpair, field)

                if field.startswith("src") or field.startswith("ref"):
                    self._convert_format(origin_file, output_file)
                else:
                    # document metadata keys
                    self._convert_meta(origin_file, field, output_file)

    def fieldnames(self, langpair):
        """
        Return a list of all the field names. For most source, this is just
        the source and the reference. For others, it might include the document
        ID for each line, or the original language (origLang).

        get_files() should return the same number of items as this.
        """
        meta = self._get_langpair_metadata(langpair)
        length = len(meta[langpair])

        assert (
            length >= 2
        ), f"Each language pair in {self.name} must have at least 2 fields."

        fields = ["src"]

        if length == 2:
            fields.append("ref")
        else:
            for i, _ in enumerate(meta[langpair][1:]):
                fields.append(f"ref:{i}")

        if not self.name.startswith("wmt08"):
            fields += ["docid", "genre", "origlang"]

        return fields


class WMTAdditionDataset(FakeSGMLDataset):
    """
    Handle special case of WMT Google addition dataset.
    """

    def _convert_format(self, input_file_path, output_filep_path):
        if input_file_path.endswith(".sgm"):
            return super()._convert_format(input_file_path, output_filep_path)
        else:
            with smart_open(input_file_path) as fin:
                with smart_open(output_filep_path, "wt") as fout:
                    for line in fin:
                        print(line.rstrip(), file=fout)



import sys

from . import DATASETS

try:
    cmd = sys.argv[1]
except IndexError:
    print(f"Usage: {sys.argv[0]} --check | --dump")
    sys.exit(1)

if cmd == "--check":
    import hashlib
    import urllib.request

    url_md5 = {}

    for item in DATASETS.values():
        if item.md5 is not None:
            assert item.data
            assert item.md5
            assert len(item.data) == len(item.md5)
            pairs = zip(item.data, item.md5)
            for url, md5_hash in pairs:
                url_md5[url] = md5_hash

    for url, md5_hash in url_md5.items():
        try:
            print("Downloading ", url)
            with urllib.request.urlopen(url) as f:
                data = f.read()
        except Exception as exc:
            raise (exc)

        if hashlib.md5(data).hexdigest() != md5_hash:
            print("MD5 check failed for", url)
elif cmd == "--dump":
    import re

    # Dumps a table in markdown format
    print(f'| {"Dataset":<30} | {"Description":<115} |')
    header = "| " + "-" * 30 + " | " + "-" * 115 + " |"
    print(header)
    for name, item in DATASETS.items():
        desc = re.sub(r"(http[s]?:\/\/\S+)", r"[URL](\1)", str(item.description))
        print(f"| {name:<30} | {desc:<115} |")



# Copyright 2017 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from collections import namedtuple
import pytest

import sacrebleu

from sacrebleu.metrics import BLEU


EPSILON = 1e-8

Statistics = namedtuple('Statistics', ['common', 'total'])

test_raw_bleu_cases = [
    # This now returns 0.0 score (#141)
    (["this is a test", "another test"], [["ref1", "ref2"]], 0.0),
    (["this is a test"], [["this is a test"]], 1.0),
    (["this is a fest"], [["this is a test"]], 0.223606797749979)]

# test for README example with empty hypothesis strings check
_refs = [
    ['The dog bit the man.', 'It was not unexpected.', 'The man bit him first.'],
    ['The dog had bit the man.', 'No one was surprised.', 'The man had bitten the dog.'],
]

_hyps = [
    'The dog bit the man.',
    "It wasn't surprising.",
    'The man had just bitten him.',
]

test_corpus_bleu_cases = [
    (_hyps, _refs, {}, 48.530827),   # test for default BLEU settings
    (('', '', ''), _refs, {}, 0.0),  # ensure that empty hypotheses are not removed
    (_hyps, _refs, {'tokenize': 'none'}, 49.1919566),
    (_hyps, _refs, {'tokenize': '13a'}, 48.530827),
    (_hyps, _refs, {'tokenize': 'intl'}, 43.91623493),
    (_hyps, _refs, {'smooth_method': 'none'}, 48.530827),
]

test_case_offset = [(["am I am a character sequence"], [["I am a symbol string sequence a a"]], 0.1555722182, 0)]

# statistic structure:
# - common counts
# - total counts
# - hyp_count
# - ref_count

test_case_statistics = [(["am I am a character sequence"], [["I am a symbol string sequence a a"]],
                         Statistics([4, 2, 1, 0], [6, 5, 4, 3]))]

test_case_scoring = [((Statistics([9, 7, 5, 3], [10, 8, 6, 4]), 11, 11), 0.8375922397)]

test_case_effective_order = [(["test"], [["a test"]], 0.3678794411714425),
                             (["a test"], [["a test"]], 1.0),
                             (["a little test"], [["a test"]], 0.03218297948685433)]


# testing that right score is returned for null statistics and different offsets
# format: stat, offset, expected score
test_case_degenerate_stats = [((Statistics([0, 0, 0, 0], [4, 4, 2, 1]), 0, 1), 0.0, 0.0),
                              ((Statistics([0, 0, 0, 0], [10, 11, 12, 0]), 14, 10), 0.0, 0.0),
                              ((Statistics([0, 0, 0, 0], [0, 0, 0, 0]), 0, 0), 0.0, 0.0),
                              ((Statistics([6, 5, 4, 0], [6, 5, 4, 3]), 6, 6), 0.0, 0.0),
                              ((Statistics([0, 0, 0, 0], [0, 0, 0, 0]), 0, 0), 0.1, 0.0),
                              ((Statistics([0, 0, 0, 0], [0, 0, 0, 0]), 1, 5), 0.01, 0.0)]


@pytest.mark.parametrize("hypotheses, references, expected_bleu", test_raw_bleu_cases)
def test_raw_bleu(hypotheses, references, expected_bleu):
    bleu = sacrebleu.raw_corpus_bleu(hypotheses, references, .01).score / 100
    assert abs(bleu - expected_bleu) < EPSILON


@pytest.mark.parametrize("hypotheses, references, kwargs, expected_bleu", test_corpus_bleu_cases)
def test_corpus_bleu(hypotheses, references, kwargs, expected_bleu):
    bleu = sacrebleu.corpus_bleu(hypotheses, references, **kwargs).score
    assert abs(bleu - expected_bleu) < EPSILON


@pytest.mark.parametrize("hypotheses, references, expected_bleu", test_case_effective_order)
def test_effective_order(hypotheses, references, expected_bleu):
    bleu = sacrebleu.raw_corpus_bleu(hypotheses, references, .01).score / 100
    assert abs(bleu - expected_bleu) < EPSILON


@pytest.mark.parametrize("hypothesis, reference, expected_stat", test_case_statistics)
def test_statistics(hypothesis, reference, expected_stat):
    result = sacrebleu.raw_corpus_bleu(hypothesis, reference, .01)
    stat = Statistics(result.counts, result.totals)
    assert stat == expected_stat


@pytest.mark.parametrize("statistics, expected_score", test_case_scoring)
def test_scoring(statistics, expected_score):
    score = BLEU.compute_bleu(statistics[0].common, statistics[0].total, statistics[1], statistics[2]).score / 100
    assert abs(score - expected_score) < EPSILON


@pytest.mark.parametrize("hypothesis, reference, expected_with_offset, expected_without_offset",
                         test_case_offset)
def test_offset(hypothesis, reference, expected_with_offset, expected_without_offset):
    score_without_offset = sacrebleu.raw_corpus_bleu(hypothesis, reference, 0.0).score / 100
    assert abs(expected_without_offset - score_without_offset) < EPSILON

    # let it use BLEU's internal default of 0.1 through passing `None`
    score_with_offset = sacrebleu.raw_corpus_bleu(hypothesis, reference, None).score / 100
    assert abs(expected_with_offset - score_with_offset) < EPSILON

    # let it use BLEU's internal default of 0.1
    score_with_offset = sacrebleu.raw_corpus_bleu(hypothesis, reference).score / 100
    assert abs(expected_with_offset - score_with_offset) < EPSILON


@pytest.mark.parametrize("statistics, offset, expected_score", test_case_degenerate_stats)
def test_degenerate_statistics(statistics, offset, expected_score):
    score = BLEU.compute_bleu(
        statistics[0].common,
        statistics[0].total,
        statistics[1],
        statistics[2],
        smooth_method='floor', smooth_value=offset).score / 100
    assert score == expected_score


test_bleu_max_order = [
    (1, _hyps, _refs, "77.65"),
    (2, _hyps, _refs, "60.50"),
    (3, _hyps, _refs, "53.93"),
    (4, _hyps, _refs, "48.53"),
    (5, _hyps, _refs, "46.14"),
    (6, _hyps, _refs, "43.28"),
]


@pytest.mark.parametrize("order, hyps, refs, expected_bleu", test_bleu_max_order)
def test_max_ngram_order(order, hyps, refs, expected_bleu):
    bleu = BLEU(max_ngram_order=order).corpus_score(hyps, refs)
    assert f"{bleu.score:.2f}" == expected_bleu



from .tokenizer_base import BaseTokenizer

class NoneTokenizer(BaseTokenizer):
    """Don't apply any tokenization. Not recommended!."""

    def signature(self):
        return 'none'

    def __call__(self, line):
        return line



"""This module implements various utility functions for the TER metric."""

# Copyright 2020 Memsource
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import math
from typing import List, Tuple, Dict


_COST_INS = 1
_COST_DEL = 1
_COST_SUB = 1

# Tercom-inspired limits
_MAX_SHIFT_SIZE = 10
_MAX_SHIFT_DIST = 50
_BEAM_WIDTH = 25

# Our own limits
_MAX_CACHE_SIZE = 10000
_MAX_SHIFT_CANDIDATES = 1000
_INT_INFINITY = int(1e16)

_OP_INS = 'i'
_OP_DEL = 'd'
_OP_NOP = ' '
_OP_SUB = 's'
_OP_UNDEF = 'x'

_FLIP_OPS = str.maketrans(_OP_INS + _OP_DEL, _OP_DEL + _OP_INS)


def translation_edit_rate(words_hyp: List[str], words_ref: List[str]) -> Tuple[int, int]:
    """Calculate the translation edit rate.

    :param words_hyp: Tokenized translation hypothesis.
    :param words_ref: Tokenized reference translation.
    :return: tuple (number of edits, length)
    """
    n_words_ref = len(words_ref)
    n_words_hyp = len(words_hyp)
    if n_words_ref == 0:
        # FIXME: This trace here is not used?
        trace = _OP_DEL * n_words_hyp
        # special treatment of empty refs
        return n_words_hyp, 0

    cached_ed = BeamEditDistance(words_ref)
    shifts = 0

    input_words = words_hyp
    checked_candidates = 0
    while True:
        # do shifts until they stop reducing the edit distance
        delta, new_input_words, checked_candidates = _shift(
            input_words, words_ref, cached_ed, checked_candidates)

        if checked_candidates >= _MAX_SHIFT_CANDIDATES:
            break

        if delta <= 0:
            break
        shifts += 1
        input_words = new_input_words

    edit_distance, trace = cached_ed(input_words)
    total_edits = shifts + edit_distance

    return total_edits, n_words_ref


def _shift(words_h: List[str], words_r: List[str], cached_ed,
           checked_candidates: int) -> Tuple[int, List[str], int]:
    """Attempt to shift words in hypothesis to match reference.

    Returns the shift that reduces the edit distance the most.

    Note that the filtering of possible shifts and shift selection are heavily
    based on somewhat arbitrary heuristics. The code here follows as closely
    as possible the logic in Tercom, not always justifying the particular design
    choices.

    :param words_h: Hypothesis.
    :param words_r: Reference.
    :param cached_ed: Cached edit distance.
    :param checked_candidates: Number of shift candidates that were already
                               evaluated.
    :return: (score, shifted_words, checked_candidates). Best shift and updated
             number of evaluated shift candidates.
    """
    pre_score, inv_trace = cached_ed(words_h)

    # to get alignment, we pretend we are rewriting reference into hypothesis,
    # so we need to flip the trace of edit operations
    trace = _flip_trace(inv_trace)
    align, ref_err, hyp_err = trace_to_alignment(trace)

    best = None

    for start_h, start_r, length in _find_shifted_pairs(words_h, words_r):
        # don't do the shift unless both the hypothesis was wrong and the
        # reference doesn't match hypothesis at the target position
        if sum(hyp_err[start_h: start_h + length]) == 0:
            continue

        if sum(ref_err[start_r: start_r + length]) == 0:
            continue

        # don't try to shift within the subsequence
        if start_h <= align[start_r] < start_h + length:
            continue

        prev_idx = -1
        for offset in range(-1, length):
            if start_r + offset == -1:
                idx = 0  # insert before the beginning
            elif start_r + offset in align:
                # Unlike Tercom which inserts *after* the index, we insert
                # *before* the index.
                idx = align[start_r + offset] + 1
            else:
                break  # offset is out of bounds => aims past reference

            if idx == prev_idx:
                continue  # skip idx if already tried

            prev_idx = idx

            shifted_words = _perform_shift(words_h, start_h, length, idx)
            assert(len(shifted_words) == len(words_h))

            # Elements of the tuple are designed to replicate Tercom ranking
            # of shifts:
            candidate = (
                pre_score - cached_ed(shifted_words)[0],  # highest score first
                length,  # then, longest match first
                -start_h,  # then, earliest match first
                -idx,   # then, earliest target position first
                shifted_words,
            )

            checked_candidates += 1

            if not best or candidate > best:
                best = candidate

        if checked_candidates >= _MAX_SHIFT_CANDIDATES:
            break

    if not best:
        return 0, words_h, checked_candidates
    else:
        best_score, _, _, _, shifted_words = best
        return best_score, shifted_words, checked_candidates


def _perform_shift(words: List[str], start: int, length: int, target: int) -> List[str]:
    """Perform a shift in `words` from `start` to `target`.

    :param words: Words to shift.
    :param start: Where from.
    :param length: How many words.
    :param target: Where to.
    :return: Shifted words.
    """
    if target < start:
        # shift before previous position
        return words[:target] + words[start: start + length] \
            + words[target: start] + words[start + length:]
    elif target > start + length:
        # shift after previous position
        return words[:start] + words[start + length: target] \
            + words[start: start + length] + words[target:]
    else:
        # shift within the shifted string
        return words[:start] + words[start + length: length + target] \
            + words[start: start + length] + words[length + target:]


def _find_shifted_pairs(words_h: List[str], words_r: List[str]):
    """Find matching word sub-sequences in two lists of words.

    Ignores sub-sequences starting at the same position.

    :param words_h: First word list.
    :param words_r: Second word list.
    :return: Yields tuples of (h_start, r_start, length) such that:
         words_h[h_start:h_start+length] = words_r[r_start:r_start+length]
    """
    n_words_h = len(words_h)
    n_words_r = len(words_r)
    for start_h in range(n_words_h):
        for start_r in range(n_words_r):
            # this is slightly different from what tercom does but this should
            # really only kick in in degenerate cases
            if abs(start_r - start_h) > _MAX_SHIFT_DIST:
                continue

            length = 0
            while words_h[start_h + length] == words_r[start_r + length] and length < _MAX_SHIFT_SIZE:
                length += 1

                yield start_h, start_r, length

                # If one sequence is consumed, stop processing
                if n_words_h == start_h + length or n_words_r == start_r + length:
                    break


def _flip_trace(trace):
    """Flip the trace of edit operations.

    Instead of rewriting a->b, get a recipe for rewriting b->a.

    Simply flips insertions and deletions.
    """
    return trace.translate(_FLIP_OPS)


def trace_to_alignment(trace: str) -> Tuple[Dict, List, List]:
    """Transform trace of edit operations into an alignment of the sequences.

    :param trace: Trace of edit operations (' '=no change or 's'/'i'/'d').
    :return: Alignment, error positions in reference, error positions in hypothesis.
    """
    pos_hyp = -1
    pos_ref = -1
    hyp_err = []
    ref_err = []
    align = {}

    # we are rewriting a into b
    for op in trace:
        if op == _OP_NOP:
            pos_hyp += 1
            pos_ref += 1
            align[pos_ref] = pos_hyp
            hyp_err.append(0)
            ref_err.append(0)
        elif op == _OP_SUB:
            pos_hyp += 1
            pos_ref += 1
            align[pos_ref] = pos_hyp
            hyp_err.append(1)
            ref_err.append(1)
        elif op == _OP_INS:
            pos_hyp += 1
            hyp_err.append(1)
        elif op == _OP_DEL:
            pos_ref += 1
            align[pos_ref] = pos_hyp
            ref_err.append(1)
        else:
            raise Exception(f"unknown operation {op!r}")

    return align, ref_err, hyp_err


class BeamEditDistance:
    """Edit distance with several features required for TER calculation.

        * internal cache
        * "beam" search
        * tracking of edit operations

    The internal self._cache works like this:

    Keys are words of the hypothesis. Values are tuples (next_node, row) where:

        * next_node is the cache for the next word in the sequence
        * row is the stored row of the edit distance matrix

    Effectively, caching allows to skip several rows in the edit distance
    matrix calculation and instead, to initialize the computation with the last
    matching matrix row.

    Beam search, as implemented here, only explores a fixed-size sub-row of
    candidates around the matrix diagonal (more precisely, it's a
    "pseudo"-diagonal since we take the ratio of sequence lengths into account).

    Tracking allows to reconstruct the optimal sequence of edit operations.

    :param words_ref: A list of reference tokens.
    """
    def __init__(self, words_ref: List[str]):
        """`BeamEditDistance` initializer."""
        self._words_ref = words_ref
        self._n_words_ref = len(self._words_ref)

        # first row corresponds to insertion operations of the reference,
        # so we do 1 edit operation per reference word
        self._initial_row = [(i * _COST_INS, _OP_INS)
                             for i in range(self._n_words_ref + 1)]

        self._cache = {}  # type: Dict[str, Tuple]
        self._cache_size = 0

        # Precomputed empty matrix row. Contains infinities so that beam search
        # avoids using the uninitialized cells.
        self._empty_row = [(_INT_INFINITY, _OP_UNDEF)] * (self._n_words_ref + 1)

    def __call__(self, words_hyp: List[str]) -> Tuple[int, str]:
        """Calculate edit distance between self._words_ref and the hypothesis.

        Uses cache to skip some of the computation.

        :param words_hyp: Words in translation hypothesis.
        :return: Edit distance score.
        """

        # skip initial words in the hypothesis for which we already know the
        # edit distance
        start_position, dist = self._find_cache(words_hyp)

        # calculate the rest of the edit distance matrix
        edit_distance, newly_created_matrix, trace = self._edit_distance(
            words_hyp, start_position, dist)

        # update our cache with the newly calculated rows
        self._add_cache(words_hyp, newly_created_matrix)

        return edit_distance, trace

    def _edit_distance(self, words_h: List[str], start_h: int,
                       cache: List[List[Tuple[int, str]]]) -> Tuple[int, List, str]:
        """Actual edit distance calculation.

        Can be initialized with the last cached row and a start position in
        the hypothesis that it corresponds to.

        :param words_h: Words in translation hypothesis.
        :param start_h: Position from which to start the calculation.
                        (This is zero if no cache match was found.)
        :param cache: Precomputed rows corresponding to edit distance matrix
                      before `start_h`.
        :return: Edit distance value, newly computed rows to update the
                 cache, trace.
        """

        n_words_h = len(words_h)

        # initialize the rest of the matrix with infinite edit distances
        rest_empty = [list(self._empty_row)
                      for _ in range(n_words_h - start_h)]

        dist = cache + rest_empty

        assert len(dist) == n_words_h + 1

        length_ratio = self._n_words_ref / n_words_h if words_h else 1

        # in some crazy sentences, the difference in length is so large that
        # we may end up with zero overlap with previous row
        if _BEAM_WIDTH < length_ratio / 2:
            beam_width = math.ceil(length_ratio / 2 + _BEAM_WIDTH)
        else:
            beam_width = _BEAM_WIDTH

        # calculate the Levenshtein distance
        for i in range(start_h + 1, n_words_h + 1):
            pseudo_diag = math.floor(i * length_ratio)
            min_j = max(0, pseudo_diag - beam_width)
            max_j = min(self._n_words_ref + 1, pseudo_diag + beam_width)

            if i == n_words_h:
                max_j = self._n_words_ref + 1

            for j in range(min_j, max_j):
                if j == 0:
                    dist[i][j] = (dist[i - 1][j][0] + _COST_DEL, _OP_DEL)
                else:
                    if words_h[i - 1] == self._words_ref[j - 1]:
                        cost_sub = 0
                        op_sub = _OP_NOP
                    else:
                        cost_sub = _COST_SUB
                        op_sub = _OP_SUB

                    # Tercom prefers no-op/sub, then insertion, then deletion.
                    # But since we flip the trace and compute the alignment from
                    # the inverse, we need to swap order of insertion and
                    # deletion in the preference.
                    ops = (
                        (dist[i - 1][j - 1][0] + cost_sub, op_sub),
                        (dist[i - 1][j][0] + _COST_DEL, _OP_DEL),
                        (dist[i][j - 1][0] + _COST_INS, _OP_INS),
                    )

                    for op_cost, op_name in ops:
                        if dist[i][j][0] > op_cost:
                            dist[i][j] = op_cost, op_name

        # get the trace
        trace = ""
        i = n_words_h
        j = self._n_words_ref

        while i > 0 or j > 0:
            op = dist[i][j][1]
            trace = op + trace
            if op in (_OP_SUB, _OP_NOP):
                i -= 1
                j -= 1
            elif op == _OP_INS:
                j -= 1
            elif op == _OP_DEL:
                i -= 1
            else:
                raise Exception(f"unknown operation {op!r}")

        return dist[-1][-1][0], dist[len(cache):], trace

    def _add_cache(self, words_hyp: List[str], mat: List[List[Tuple]]):
        """Add newly computed rows to cache.

        Since edit distance is only calculated on the hypothesis suffix that
        was not in cache, the number of rows in `mat` may be shorter than
        hypothesis length. In that case, we skip over these initial words.

        :param words_hyp: Hypothesis words.
        :param mat: Edit distance matrix rows for each position.
        """
        if self._cache_size >= _MAX_CACHE_SIZE:
            return

        node = self._cache

        n_mat = len(mat)

        # how many initial words to skip
        skip_num = len(words_hyp) - n_mat

        # jump through the cache to the current position
        for i in range(skip_num):
            node = node[words_hyp[i]][0]

        assert len(words_hyp[skip_num:]) == n_mat

        # update cache with newly computed rows
        for word, row in zip(words_hyp[skip_num:], mat):
            if word not in node:
                node[word] = ({}, tuple(row))
                self._cache_size += 1
            value = node[word]
            node = value[0]

    def _find_cache(self, words_hyp: List[str]) -> Tuple[int, List[List]]:
        """Find the already computed rows of the edit distance matrix in cache.

        Returns a partially computed edit distance matrix.

        :param words_hyp: Translation hypothesis.
        :return: Tuple (start position, dist).
        """
        node = self._cache
        start_position = 0
        dist = [self._initial_row]
        for word in words_hyp:
            if word in node:
                start_position += 1
                node, row = node[word]
                dist.append(row)
            else:
                break

        return start_position, dist



"""The implementation of various metrics."""

from .bleu import BLEU, BLEUScore   # noqa: F401
from .chrf import CHRF, CHRFScore   # noqa: F401
from .ter import TER, TERScore      # noqa: F401

METRICS = {
    'BLEU': BLEU,
    'CHRF': CHRF,
    'TER': TER,
}



#!/usr/bin/env python
import sys
import time
import statistics

sys.path.insert(0, '.')

import sacrebleu  # noqa: E402
from sacrebleu.metrics import BLEU, CHRF  # noqa: E402


N_REPEATS = 5


sys_files = [
    'data/wmt17-submitted-data/txt/system-outputs/newstest2017/cs-en/newstest2017.PJATK.4760.cs-en',
    'data/wmt17-submitted-data/txt/system-outputs/newstest2017/cs-en/newstest2017.uedin-nmt.4955.cs-en',
    'data/wmt17-submitted-data/txt/system-outputs/newstest2017/cs-en/newstest2017.online-A.0.cs-en',
    'data/wmt17-submitted-data/txt/system-outputs/newstest2017/cs-en/newstest2017.online-B.0.cs-en',
]

ref_files = ['data/wmt17-submitted-data/txt/references/newstest2017-csen-ref.en']

metrics = [
    # BLEU
    (BLEU, {}),
    (BLEU, {'tokenize': 'intl'}),
    (BLEU, {'tokenize': 'none', 'force': True}),
    # CHRF
    (CHRF, {}),
    (CHRF, {'whitespace': True}),
    # CHRF++
    (CHRF, {'word_order': 2}),
]


def create_metric(klass, kwargs, refs=None):
    if refs:
        # caching mode
        kwargs['references'] = refs
    return klass(**kwargs)


def read_files(*args):
    lines = []
    for fname in args:
        cur_lines = []
        with open(fname) as f:
            for line in f:
                cur_lines.append(line.strip())
        lines.append(cur_lines)
    return lines


def measure(metric_klass, metric_kwargs, systems, refs, cache=False):
    scores = []
    durations = []

    if cache:
        # caching mode
        metric_kwargs['references'] = refs
    st = time.time()
    metric = metric_klass(**metric_kwargs)

    for system in systems:
        sc = metric.corpus_score(system, None if cache else refs).score
        dur = time.time() - st
        print(f'{dur:.3f}', end=' ')
        durations.append(dur)
        scores.append(sc)
        st = time.time()

    durations = sorted(durations)
    median = durations[len(durations) // 2]
    std = statistics.pstdev(durations)
    mean = sum(durations) / len(durations)
    print(f' || mean: {mean:.3f} -- median: {median:.3f} -- stdev: {std:.3f}')


if __name__ == '__main__':

    systems = read_files(*sys_files)
    refs = read_files(*ref_files)

    msg = f'SacreBLEU {sacrebleu.__version__} performance tests'
    print('-' * len(msg) + '\n' + msg + '\n' + '-' * len(msg))

    for klass, kwargs in metrics:
        print(klass.__name__, kwargs)

        print(' > [no-cache] ', end='')
        measure(klass, kwargs, systems, refs, cache=False)

        print(' >   [cached] ', end='')
        measure(klass, kwargs, systems, refs, cache=True)



import pytest
import sacrebleu

EPSILON = 1e-3

test_cases = [
    (['aaaa bbbb cccc dddd'], ['aaaa bbbb cccc dddd'], 0),  # perfect match
    (['dddd eeee ffff'], ['aaaa bbbb cccc'], 1),  # no overlap
    ([''], ['a'], 1),  # corner case, empty hypothesis
    (['a'], [''], 1),  # corner case, empty reference
    ([''], [''], 0),  # corner case, both reference and hypothesis empty, we define it as 0.0
    (['d e f g h a b c'], ['a b c d e f g h'], 1 / 8),  # a single shift fixes MT
    (
        [
            'wählen Sie " Bild neu berechnen , " um beim Ändern der Bildgröße Pixel hinzuzufügen oder zu entfernen , damit das Bild ungefähr dieselbe Größe aufweist wie die andere Größe .',
            'wenn Sie alle Aufgaben im aktuellen Dokument aktualisieren möchten , wählen Sie im Menü des Aufgabenbedienfelds die Option " Alle Aufgaben aktualisieren . "',
            'klicken Sie auf der Registerkarte " Optionen " auf die Schaltfläche " Benutzerdefiniert " und geben Sie Werte für " Fehlerkorrektur-Level " und " Y / X-Verhältnis " ein .',
            'Sie können beispielsweise ein Dokument erstellen , das ein Auto über die Bühne enthält .',
            'wählen Sie im Dialogfeld " Neu aus Vorlage " eine Vorlage aus und klicken Sie auf " Neu . "',
        ],
        [
            'wählen Sie " Bild neu berechnen , " um beim Ändern der Bildgröße Pixel hinzuzufügen oder zu entfernen , damit die Darstellung des Bildes in einer anderen Größe beibehalten wird .',
            'wenn Sie alle Aufgaben im aktuellen Dokument aktualisieren möchten , wählen Sie im Menü des Aufgabenbedienfelds die Option " Alle Aufgaben aktualisieren . "',
            'klicken Sie auf der Registerkarte " Optionen " auf die Schaltfläche " Benutzerdefiniert " und geben Sie für " Fehlerkorrektur-Level " und " Y / X-Verhältnis " niedrigere Werte ein .',
            'Sie können beispielsweise ein Dokument erstellen , das ein Auto enthalt , das sich über die Bühne bewegt .',
            'wählen Sie im Dialogfeld " Neu aus Vorlage " eine Vorlage aus und klicken Sie auf " Neu . "',
        ],
        0.136  # realistic example from WMT dev data (2019)
    ),
]


@pytest.mark.parametrize("hypotheses, references, expected_score", test_cases)
def test_ter(hypotheses, references, expected_score):
    metric = sacrebleu.metrics.TER()
    score = metric.corpus_score(hypotheses, [references]).score
    assert abs(score - 100 * expected_score) < EPSILON



# Copyright 2017--2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

##############

# MIT License
# Copyright (c) 2017 - Shujian Huang <huangsj@nju.edu.cn>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Author: Shujian Huang huangsj@nju.edu.cn


from functools import lru_cache

from .tokenizer_base import BaseTokenizer
from .tokenizer_re import TokenizerRegexp

_UCODE_RANGES = [
    (u'\u3400', u'\u4db5'),  # CJK Unified Ideographs Extension A, release 3.0
    (u'\u4e00', u'\u9fa5'),  # CJK Unified Ideographs, release 1.1
    (u'\u9fa6', u'\u9fbb'),  # CJK Unified Ideographs, release 4.1
    (u'\uf900', u'\ufa2d'),  # CJK Compatibility Ideographs, release 1.1
    (u'\ufa30', u'\ufa6a'),  # CJK Compatibility Ideographs, release 3.2
    (u'\ufa70', u'\ufad9'),  # CJK Compatibility Ideographs, release 4.1
    (u'\u20000', u'\u2a6d6'),  # (UTF16) CJK Unified Ideographs Extension B, release 3.1
    (u'\u2f800', u'\u2fa1d'),  # (UTF16) CJK Compatibility Supplement, release 3.1
    (u'\uff00', u'\uffef'),  # Full width ASCII, full width of English punctuation,
                             # half width Katakana, half wide half width kana, Korean alphabet
    (u'\u2e80', u'\u2eff'),  # CJK Radicals Supplement
    (u'\u3000', u'\u303f'),  # CJK punctuation mark
    (u'\u31c0', u'\u31ef'),  # CJK stroke
    (u'\u2f00', u'\u2fdf'),  # Kangxi Radicals
    (u'\u2ff0', u'\u2fff'),  # Chinese character structure
    (u'\u3100', u'\u312f'),  # Phonetic symbols
    (u'\u31a0', u'\u31bf'),  # Phonetic symbols (Taiwanese and Hakka expansion)
    (u'\ufe10', u'\ufe1f'),
    (u'\ufe30', u'\ufe4f'),
    (u'\u2600', u'\u26ff'),
    (u'\u2700', u'\u27bf'),
    (u'\u3200', u'\u32ff'),
    (u'\u3300', u'\u33ff'),
]


class TokenizerZh(BaseTokenizer):

    def signature(self):
        return 'zh'

    def __init__(self):
        self._post_tokenizer = TokenizerRegexp()

    @staticmethod
    @lru_cache(maxsize=2**16)
    def _is_chinese_char(uchar):
        """
        :param uchar: input char in unicode
        :return: whether the input char is a Chinese character.
        """
        for start, end in _UCODE_RANGES:
            if start <= uchar <= end:
                return True
        return False

    @lru_cache(maxsize=2**16)
    def __call__(self, line):
        """The tokenization of Chinese text in this script contains two
        steps: separate each Chinese characters (by utf-8 encoding); tokenize
        the non Chinese part (following the `13a` i.e. mteval tokenizer).

        Author: Shujian Huang huangsj@nju.edu.cn

        :param line: input sentence
        :return: tokenized sentence
        """

        line = line.strip()
        line_in_chars = ""

        # TODO: the below code could probably be replaced with the following:
        # @ozan: Gives slightly different scores, need to investigate
        # import regex
        # line = regex.sub(r'(\p{Han})', r' \1 ', line)
        for char in line:
            if self._is_chinese_char(char):
                line_in_chars += " "
                line_in_chars += char
                line_in_chars += " "
            else:
                line_in_chars += char

        return self._post_tokenizer(line_in_chars)



"""
The base class for all types of datasets.
"""
import os
import re
from abc import ABCMeta, abstractmethod
from typing import Dict, List, Optional

from ..utils import SACREBLEU_DIR, download_file, smart_open


class Dataset(metaclass=ABCMeta):
    def __init__(
        self,
        name: str,
        data: Optional[List[str]] = None,
        description: Optional[str] = None,
        citation: Optional[str] = None,
        md5: Optional[List[str]] = None,
        langpairs=Dict[str, List[str]],
        **kwargs,
    ):
        """
        Params come from the values in DATASETS.

        :param name: Name of the dataset.
        :param data: URL of the raw data of the dataset.
        :param description: Description of the dataset.
        :param citation: Citation for the dataset.
        :param md5: MD5 checksum of the dataset.
        :param langpairs: List of available language pairs.
        """
        self.name = name
        self.data = data
        self.description = description
        self.citation = citation
        self.md5 = md5
        self.langpairs = langpairs
        self.kwargs = kwargs

        # Don't do any downloading or further processing now.
        # Only do that lazily, when asked.

        # where to store the dataset
        self._outdir = os.path.join(SACREBLEU_DIR, self.name)
        self._rawdir = os.path.join(self._outdir, "raw")

    def maybe_download(self):
        """
        If the dataset isn't downloaded, use utils/download_file()
        This can be implemented here in the base class. It should write
        to ~/.sacreleu/DATASET/raw exactly as it does now.
        """
        os.makedirs(self._rawdir, exist_ok=True)

        expected_checksums = self.md5 if self.md5 else [None] * len(self.data)

        for url, expected_md5 in zip(self.data, expected_checksums):
            tarball = os.path.join(self._rawdir, self._get_tarball_filename(url))

            download_file(
                url, tarball, extract_to=self._rawdir, expected_md5=expected_md5
            )

    @staticmethod
    def _clean(s):
        """
        Removes trailing and leading spaces and collapses multiple consecutive internal spaces to a single one.

        :param s: The string.
        :return: A cleaned-up string.
        """
        return re.sub(r"\s+", " ", s.strip())

    def _get_tarball_filename(self, url):
        """
        Produces a local filename for tarball.
        :param url: The url to download.
        :return: A name produced from the dataset identifier and the URL basename.
        """
        return self.name.replace("/", "_") + "." + os.path.basename(url)

    def _get_txt_file_path(self, langpair, fieldname):
        """
        Given the language pair and fieldname, return the path to the text file.
        The format is: ~/.sacrebleu/DATASET/DATASET.LANGPAIR.FIELDNAME

        :param langpair: The language pair.
        :param fieldname: The fieldname.
        :return: The path to the text file.
        """
        # handle the special case of subsets. e.g. "wmt21/dev" > "wmt21_dev"
        name = self.name.replace("/", "_")
        # Colons are used to distinguish multiple references, but are not supported in Windows filenames
        fieldname = fieldname.replace(":", "-")
        return os.path.join(self._outdir, f"{name}.{langpair}.{fieldname}")

    def _get_langpair_metadata(self, langpair):
        """
        Given a language pair, return the metadata for that language pair.
        Deal with errors if the language pair is not available.

        :param langpair: The language pair. e.g. "en-de"
        :return: Dict format which is same as self.langpairs.
        """
        if langpair is None:
            langpairs = self.langpairs
        elif langpair not in self.langpairs:
            raise Exception(f"No such language pair {self.name}/{langpair}")
        else:
            langpairs = {langpair: self.langpairs[langpair]}

        return langpairs

    @abstractmethod
    def process_to_text(self, langpair=None) -> None:
        """Processes raw files to plain text files.

        :param langpair: The language pair to process. e.g. "en-de". If None, all files will be processed.
        """
        pass

    def fieldnames(self, langpair) -> List[str]:
        """
        Return a list of all the field names. For most source, this is just
        the source and the reference. For others, it might include the document
        ID for each line, or the original language (origLang).

        get_files() should return the same number of items as this.

        :param langpair: The language pair (e.g., "de-en")
        :return: a list of field names
        """
        return ["src", "ref"]

    def __iter__(self, langpair):
        """
        Iterates over all fields (source, references, and other metadata) defined
        by the dataset.
        """
        all_files = self.get_files(langpair)
        all_fins = [smart_open(f) for f in all_files]

        for item in zip(*all_fins):
            yield item

    def source(self, langpair):
        """
        Return an iterable over the source lines.
        """
        source_file = self.get_source_file(langpair)
        with smart_open(source_file) as fin:
            for line in fin:
                yield line.strip()

    def references(self, langpair):
        """
        Return an iterable over the references.
        """
        ref_files = self.get_reference_files(langpair)
        ref_fins = [smart_open(f) for f in ref_files]

        for item in zip(*ref_fins):
            yield item

    def get_source_file(self, langpair):
        all_files = self.get_files(langpair)
        all_fields = self.fieldnames(langpair)
        index = all_fields.index("src")
        return all_files[index]

    def get_reference_files(self, langpair):
        all_files = self.get_files(langpair)
        all_fields = self.fieldnames(langpair)
        ref_files = [
            f for f, field in zip(all_files, all_fields) if field.startswith("ref")
        ]
        return ref_files

    def get_files(self, langpair):
        """
        Returns the path of the source file and all reference files for
        the provided test set / language pair.
        Downloads the references first if they are not already local.

        :param langpair: The language pair (e.g., "de-en")
        :return: a list of the source file and all reference files
        """
        fields = self.fieldnames(langpair)
        files = [self._get_txt_file_path(langpair, field) for field in fields]

        for file in files:
            if not os.path.exists(file):
                self.process_to_text(langpair)
        return files



from .fake_sgml import FakeSGMLDataset


class IWSLTXMLDataset(FakeSGMLDataset):
    """IWSLT dataset format. Can be parsed with the lxml parser."""

    # Same as FakeSGMLDataset. Nothing to do here.
    pass



import os
import shutil
import random

import sacrebleu.dataset as dataset
from sacrebleu.utils import smart_open


def test_maybe_download():
    """
    Test the maybe_download function in Dataset class.

    Check a few random datasets for downloading and correct file placement.
    """
    # ensure all file have been downloaded
    selected_datasets = random.choices(list(dataset.DATASETS.values()), k=10)
    for ds in selected_datasets:
        shutil.rmtree(ds._rawdir, ignore_errors=True)
        ds.maybe_download()

        all_files = os.listdir(ds._rawdir)
        for url in ds.data:
            filename = ds._get_tarball_filename(url)
            assert filename in all_files
            filepath = os.path.join(ds._rawdir, filename)
            assert os.path.getsize(filepath) > 0


def test_process_to_text():
    """
    Test the function `process_to_text` in Dataset class.

    Ensure each field of specified language pair have the same length.
    """
    selected_datasets = random.choices(list(dataset.DATASETS.values()), k=10)
    for ds in selected_datasets:
        if os.path.exists(ds._outdir):
            for filename in os.listdir(ds._outdir):
                filepath = os.path.join(ds._outdir, filename)
                if os.path.isfile(filepath):
                    os.remove(filepath)

        ds.process_to_text()

        for pair in ds.langpairs:
            all_files = ds.get_files(pair)

            # count the number of lines in each file
            num_lines = [sum(1 for _ in smart_open(f)) for f in all_files]

            # ensure no empty file
            assert num_lines[0] > 0

            # assert each file has the same length
            assert all(x == num_lines[0] for x in num_lines)


def test_get_files_and_fieldnames():
    """
    Test the functions `get_files` and `fieldnames` in Dataset class.

    Ensure the length of the returned list is correct.
    `get_files()` should return the same number of items as `fieldnames()`.
    """
    for ds in dataset.DATASETS.values():
        for pair in ds.langpairs:
            assert len(ds.get_files(pair)) == len(ds.fieldnames(pair))


def test_source_and_references():
    """
    Test the functions `source` and `references` in Dataset class.

    Ensure the length of source and references are equal.
    """
    for ds in dataset.DATASETS.values():
        for pair in ds.langpairs:
            src_len = len(list(ds.source(pair)))
            ref_len = len(list(ds.references(pair)))
            assert src_len == ref_len, f"source/reference failure for {ds.name}:{pair} len(source)={src_len} len(references)={ref_len}"


def test_wmt22_references():
    """
    WMT21 added the ability to specify which reference to use (among many in the XML).
    The default was "A" for everything.
    WMT22 added the ability to override this default on a per-langpair basis, by
    replacing the langpair list of paths with a dict that had the list of paths and
    the annotator override.
    """
    wmt22 = dataset.DATASETS["wmt22"]

    # make sure CS-EN returns all reference fields
    cs_en_fields = wmt22.fieldnames("cs-en")
    for ref in ["ref:B", "ref:C"]:
        assert ref in cs_en_fields
    assert "ref:A" not in cs_en_fields

    # make sure ref:B is the one used by default
    assert wmt22._get_langpair_allowed_refs("cs-en") == ["ref:B"]

    # similar check for another dataset: there should be no default ("A"),
    # and the only ref found should be the unannotated one
    assert "ref:A" not in wmt22.fieldnames("liv-en")
    assert "ref" in wmt22.fieldnames("liv-en")

    # and that ref:A is the default for all languages where it wasn't overridden
    for langpair, langpair_data in wmt22.langpairs.items():
        if isinstance(langpair_data, dict):
            assert wmt22._get_langpair_allowed_refs(langpair) != ["ref:A"]
        else:
            assert wmt22._get_langpair_allowed_refs(langpair) == ["ref:A"]





from typing import Sequence, Optional

from .metrics import BLEU, CHRF, TER, BLEUScore, CHRFScore, TERScore


######################################################################
# Backward compatibility functions for old style API access (< 1.4.11)
######################################################################
def corpus_bleu(hypotheses: Sequence[str],
                references: Sequence[Sequence[str]],
                smooth_method='exp',
                smooth_value=None,
                force=False,
                lowercase=False,
                tokenize=BLEU.TOKENIZER_DEFAULT,
                use_effective_order=False) -> BLEUScore:
    """Computes BLEU for a corpus against a single (or multiple) reference(s).
    This is the main CLI entry point for computing BLEU between a system output
    and a reference sentence.

    :param hypotheses: A sequence of hypothesis strings.
    :param references: A sequence of reference documents with document being
        defined as a sequence of reference strings.
    :param smooth_method: The smoothing method to use ('floor', 'add-k', 'exp' or 'none')
    :param smooth_value: The smoothing value for `floor` and `add-k` methods. `None` falls back to default value.
    :param force: Ignore data that looks already tokenized
    :param lowercase: Lowercase the data
    :param tokenize: The tokenizer to use
    :param use_effective_order: Don't take into account n-gram orders without any match.
    :return: a `BLEUScore` object
    """
    metric = BLEU(
        lowercase=lowercase, force=force, tokenize=tokenize,
        smooth_method=smooth_method, smooth_value=smooth_value,
        effective_order=use_effective_order)

    return metric.corpus_score(hypotheses, references)


def raw_corpus_bleu(hypotheses: Sequence[str],
                    references: Sequence[Sequence[str]],
                    smooth_value: Optional[float] = BLEU.SMOOTH_DEFAULTS['floor']) -> BLEUScore:
    """Computes BLEU for a corpus against a single (or multiple) reference(s).
    This convenience function assumes a particular set of arguments i.e.
    it disables tokenization and applies a `floor` smoothing with value `0.1`.

    This convenience call does not apply any tokenization at all,
    neither to the system output nor the reference. It just computes
    BLEU on the "raw corpus" (hence the name).

    :param hypotheses: A sequence of hypothesis strings.
    :param references: A sequence of reference documents with document being
        defined as a sequence of reference strings.
    :param smooth_value: The smoothing value for `floor`. If not given, the default of 0.1 is used.
    :return: Returns a `BLEUScore` object.

    """
    return corpus_bleu(
        hypotheses, references, smooth_method='floor',
        smooth_value=smooth_value, force=True, tokenize='none',
        use_effective_order=True)


def sentence_bleu(hypothesis: str,
                  references: Sequence[str],
                  smooth_method: str = 'exp',
                  smooth_value: Optional[float] = None,
                  lowercase: bool = False,
                  tokenize=BLEU.TOKENIZER_DEFAULT,
                  use_effective_order: bool = True) -> BLEUScore:
    """
    Computes BLEU for a single sentence against a single (or multiple) reference(s).

    Disclaimer: Computing BLEU at the sentence level is not its intended use as
    BLEU is a corpus-level metric.

    :param hypothesis: A single hypothesis string.
    :param references: A sequence of reference strings.
    :param smooth_method: The smoothing method to use ('floor', 'add-k', 'exp' or 'none')
    :param smooth_value: The smoothing value for `floor` and `add-k` methods. `None` falls back to default value.
    :param lowercase: Lowercase the data
    :param tokenize: The tokenizer to use
    :param use_effective_order: Don't take into account n-gram orders without any match.
    :return: Returns a `BLEUScore` object.
    """
    metric = BLEU(
        lowercase=lowercase, tokenize=tokenize, force=False,
        smooth_method=smooth_method, smooth_value=smooth_value,
        effective_order=use_effective_order)

    return metric.sentence_score(hypothesis, references)


def corpus_chrf(hypotheses: Sequence[str],
                references: Sequence[Sequence[str]],
                char_order: int = CHRF.CHAR_ORDER,
                word_order: int = CHRF.WORD_ORDER,
                beta: int = CHRF.BETA,
                remove_whitespace: bool = True,
                eps_smoothing: bool = False) -> CHRFScore:
    """
    Computes chrF for a corpus against a single (or multiple) reference(s).
    If `word_order` equals to 2, the metric is referred to as chrF++.

    :param hypotheses: A sequence of hypothesis strings.
    :param references: A sequence of reference documents with document being
        defined as a sequence of reference strings.
    :param char_order: Character n-gram order.
    :param word_order: Word n-gram order. If equals to 2, the metric is referred to as chrF++.
    :param beta: Determine the importance of recall w.r.t precision.
    :param eps_smoothing: If `True`, applies epsilon smoothing similar
    to reference chrF++.py, NLTK and Moses implementations. Otherwise,
    it takes into account effective match order similar to sacreBLEU < 2.0.0.
    :param remove_whitespace: If `True`, removes whitespaces prior to character n-gram extraction.
    :return: A `CHRFScore` object.
    """
    metric = CHRF(
        char_order=char_order,
        word_order=word_order,
        beta=beta,
        whitespace=not remove_whitespace,
        eps_smoothing=eps_smoothing)
    return metric.corpus_score(hypotheses, references)


def sentence_chrf(hypothesis: str,
                  references: Sequence[str],
                  char_order: int = CHRF.CHAR_ORDER,
                  word_order: int = CHRF.WORD_ORDER,
                  beta: int = CHRF.BETA,
                  remove_whitespace: bool = True,
                  eps_smoothing: bool = False) -> CHRFScore:
    """
    Computes chrF for a single sentence against a single (or multiple) reference(s).
    If `word_order` equals to 2, the metric is referred to as chrF++.

    :param hypothesis: A single hypothesis string.
    :param references: A sequence of reference strings.
    :param char_order: Character n-gram order.
    :param word_order: Word n-gram order. If equals to 2, the metric is referred to as chrF++.
    :param beta: Determine the importance of recall w.r.t precision.
    :param eps_smoothing: If `True`, applies epsilon smoothing similar
    to reference chrF++.py, NLTK and Moses implementations. Otherwise,
    it takes into account effective match order similar to sacreBLEU < 2.0.0.
    :param remove_whitespace: If `True`, removes whitespaces prior to character n-gram extraction.
    :return: A `CHRFScore` object.
    """
    metric = CHRF(
        char_order=char_order,
        word_order=word_order,
        beta=beta,
        whitespace=not remove_whitespace,
        eps_smoothing=eps_smoothing)
    return metric.sentence_score(hypothesis, references)


def corpus_ter(hypotheses: Sequence[str],
               references: Sequence[Sequence[str]],
               normalized: bool = False,
               no_punct: bool = False,
               asian_support: bool = False,
               case_sensitive: bool = False) -> TERScore:
    """
    Computes TER for a corpus against a single (or multiple) reference(s).

    :param hypotheses: A sequence of hypothesis strings.
    :param references: A sequence of reference documents with document being
        defined as a sequence of reference strings.
    :param normalized: Enable character normalization.
    :param no_punct: Remove punctuation.
    :param asian_support: Enable special treatment of Asian characters.
    :param case_sensitive: Enables case-sensitivity.
    :return: A `TERScore` object.
    """
    metric = TER(
        normalized=normalized,
        no_punct=no_punct,
        asian_support=asian_support,
        case_sensitive=case_sensitive)
    return metric.corpus_score(hypotheses, references)


def sentence_ter(hypothesis: str,
                 references: Sequence[str],
                 normalized: bool = False,
                 no_punct: bool = False,
                 asian_support: bool = False,
                 case_sensitive: bool = False) -> TERScore:
    """
    Computes TER for a single hypothesis against a single (or multiple) reference(s).

    :param hypothesis: A single hypothesis string.
    :param references: A sequence of reference strings.
    :param normalized: Enable character normalization.
    :param no_punct: Remove punctuation.
    :param asian_support: Enable special treatment of Asian characters.
    :param case_sensitive: Enable case-sensitivity.
    :return: A `TERScore` object.
    """
    metric = TER(
        normalized=normalized,
        no_punct=no_punct,
        asian_support=asian_support,
        case_sensitive=case_sensitive)
    return metric.sentence_score(hypothesis, references)
